"""
Classes to solve for Keplerian orbital parameters from different initial inputs

References:
- Shefer, V. A. (2010). New method of orbit determination from two position vectors based on solving Gauss’s equations. Solar System Research, 44, 252-266.
- Montenbruck, O., Gill, E., & Lutze, F. H. (2002). Satellite orbits: models, methods, and applications. Appl. Mech. Rev., 55(2), B27-B28.
"""


import abc
import numpy as np
from astropy.time import Time

from .orbit import Orbit
from .compute import rv
from .constants import EARTH_MU
from .utils import norm, lazy_property, newton_raphson, find_all_zeros


class TwoPosOrbitSolver(metaclass=abc.ABCMeta):
    """
    Parameters
    ----------
    r1, r2 : (3,) array_like
        Positions at t1 and t2 in meters.
    t1, t2 : float or astropy.time.Time
        Times of observations.
        If float, then should correspond to GPS seconds; i.e., seconds since
        1980-01-06 00:00:00 UTC
    mu : float, optional
        Gravitational constant of central body in m^3/s^2.  (Default: Earth's
        gravitational constant in WGS84).
    kappaSign : int, optional
        +1 for prograde, -1 for retrograde
    lam : int, optional
        Number of complete orbits between observations.  (Default: 0)
    eps : float, optional
        Iteration tolerance.
    maxiter : int, optional
        Maximum number of iterations.
    """
    def __init__(self, r1, r2, t1, t2, mu=EARTH_MU,
                 kappaSign=1, lam=0, eps=3e-16, maxiter=100):
        if isinstance(t1, Time):
            t1 = t1.gps
        if isinstance(t2, Time):
            t2 = t2.gps
        self.r1 = r1
        self.r2 = r2
        self.t1 = t1
        self.t2 = t2
        self.mu = mu
        self.kappaSign = kappaSign
        self.lam = lam
        self.eps = eps
        self.maxiter = maxiter

        # Preliminary calculations
        self.tau = self.t2 - self.t1

        self.r1mag = norm(self.r1)
        self.r2mag = norm(self.r2)
        self.U1 = self.r1 / self.r1mag
        self.U2 = self.r2 / self.r2mag

        self.r0 = self.r2 - np.dot(self.r2, self.U1) * self.U1  # M&G (2.110)
        self.r0mag = norm(self.r0)
        self.U0 = self.r0 / self.r0mag

        self.cos2f = np.dot(self.U1, self.U2)
        self.rr = np.sqrt(self.r1mag * self.r2mag)
        # Shefer (8)
        self.kappa = self.rr * norm(self.U1 + self.U2) * self.kappaSign
        self.sigma = self.rr * norm(self.U2 - self.U1)  # Shefer (8)
        self.rbar = 0.5 * (self.r1mag + self.r2mag)
        self.m = self.mu * self.tau**2 / self.kappa**3  # Shefer (5.5)
        self.ell = self.rbar / self.kappa - 0.5  # Shefer (5.5)

        # Can solve for the orbital plane immediately
        self.W = np.cross(self.U1, self.U0) * self.kappaSign
        # M&G (2.58)
        self.raan = np.arctan2(self.W[0], -self.W[1]) % (2 * np.pi)
        # M&G (2.58)
        self.i = np.arctan2(self.W[0] / np.sin(self.raan), self.W[2])

        # M&G (2.66)
        self.u1 = np.arctan2(
            self.U1[2],
            -self.U1[0] * self.W[1] + self.U1[1] * self.W[0]
        )

    def _finishOrbit(self, p):
        # M&G (2.114), good for both elliptical and hyperbolic orbits
        ecosnu1 = p / self.r1mag - 1.0
        ecosnu2 = p / self.r2mag - 1.0  # M&G (2.114)
        # M&G (2.116)
        esinnu1 = ((ecosnu1 * self.cos2f - ecosnu2)
                   / (self.r0mag / self.r2mag)
                   * self.kappaSign)
        e = np.hypot(ecosnu1, esinnu1)
        nu1 = np.arctan2(esinnu1, ecosnu1) % (2 * np.pi)  # trueAnomaly at t1
        pa = (self.u1 - nu1) % (2 * np.pi)  # M&G (2.117)
        a = (p / (1 - e**2))  # M&G (2.118)

        return Orbit.fromKeplerianElements(
            a, e, self.i, pa, self.raan, nu1, self.t1, self.mu
        )

    # The main part that varies between methods (i.e., subclasses)
    @abc.abstractmethod
    def _getP(self):
        pass

    def solve(self):
        p = self._getP()
        return self._finishOrbit(p)


class GaussTwoPosOrbitSolver(TwoPosOrbitSolver):
    """
    A class for solving two-position orbit determination problems using 
    the Gauss method. This class extends the `TwoPosOrbitSolver` base class 
    and implements a method to compute the orbital parameter `p` based on 
    Shefer's equations.

    Attributes:
        eps (float): Convergence tolerance for iterative calculations.
        maxiter (int): Maximum number of iterations allowed for convergence.
        m (float): A parameter related to the orbit determination problem.
        ell (float): A parameter related to the orbit determination problem.
        kappa (float): A constant used in orbital calculations.
        sigma (float): A constant used in orbital calculations.
        tau (float): A constant used in orbital calculations.
        mu (float): Standard gravitational parameter.

    Methods:
        _getP():
            Computes the orbital parameter `p` using an iterative approach 
            based on Shefer's equations. This method solves for `eta` 
            iteratively until convergence and then calculates `p` using the 
            converged value of `eta`.
    """
    def __init__(self, *args, **kwargs):
        TwoPosOrbitSolver.__init__(self, *args, **kwargs)

    def _getP(self):
        """Compute p from Shefer (2)."""
        # Solve for eta
        eta = 1
        d_eta = 1
        niter = 0
        while np.abs(d_eta) > self.eps and niter < self.maxiter:
            x = (self.m / eta**2 - self.ell)  # Shefer (5)
            if 1e-15 < x < 1:  # Shefer (6.5ish)
                g = 2 * np.arcsin(np.sqrt(x))
                X = (2 * g - np.sin(2 * g)) / np.sin(g)**3
            elif x < -1e-15:
                h = 2 * np.arcsinh(np.sqrt(-x))
                X = (np.sinh(2 * h) - 2 * h) / np.sinh(h)**3
            else:  # -1e-15<x<1e-15
                X = 4. / 3 + 8. / 5 * x + 64. / 35 * x
            d_eta = 1. + (self.ell + x) * X - eta
            eta += d_eta
            niter += 1
        # Plug eta into (2) to obtain p
        p = (0.5 * eta * self.kappa * self.sigma / self.tau)**2 / self.mu
        return p


class DanchickTwoPosOrbitSolver(TwoPosOrbitSolver):
    """
    A class for solving two-position orbit determination problems using 
    the Danchick method. This class extends the `TwoPosOrbitSolver` base 
    class and implements methods to compute the orbital parameter `p` 
    based on Shefer's equations.

    Attributes:
        eps (float): Convergence tolerance for iterative calculations.
        maxiter (int): Maximum number of iterations allowed for convergence.
        m (float): A parameter related to the orbit determination problem.
        ell (float): A parameter related to the orbit determination problem.
        kappa (float): A constant used in orbital calculations.
        sigma (float): A constant used in orbital calculations.
        tau (float): A constant used in orbital calculations.
        mu (float): Standard gravitational parameter.
        cos2f (float): Cosine of twice the true anomaly, used to determine 
            the branch of the solution.

    Methods:
        X(g):
            Compute the function X(g) as described in Shefer's equation (11).
            This is used in the iterative solution process.

        dXdg(g):
            Compute the derivative of X(g) with respect to g, based on 
            Shefer's equation (12). This is used to refine the solution 
            during iteration.

        _getP():
            Computes the orbital parameter `p` using iterative methods 
            based on Shefer's equations. Depending on the value of `cos2f`, 
            the method selects the appropriate branch of the solution and 
            iteratively solves for `eta` or `x` until convergence. The final 
            value of `p` is calculated using Shefer's equation (2).
    """
    def __init__(self, *args, **kwargs):
        TwoPosOrbitSolver.__init__(self, *args, **kwargs)

    @staticmethod
    def X(g):
        """Compute X(g) from Shefer (11)."""
        return (2 * g - np.sin(2 * g)) / (np.sin(g)**3)

    @staticmethod
    def dXdg(g):
        """Compute dX(g)/dg from Shefer (12)."""
        return (2 * (1 - np.cos(2 * g)) - 3 * (2 * g - np.sin(2 * g)) / np.tan(g)) / (np.sin(g)**3)

    def _getP(self):
        if self.cos2f >= 0:
            # TODO: rewrite to use newton_raphson function.
            eta = 0.5 * (np.sqrt(self.m / (self.ell + 1)) + np.sqrt(self.m / self.ell))
            Geta = 1
            niter = 0
            while np.abs(Geta) > self.eps and niter < self.maxiter:
                x = self.m / eta**2 - self.ell  # Shefer (14)
                if (1 - 2 * x) > 1 or (1 - 2 * x) < -1:
                    raise RuntimeError("Invalid x")
                g = np.arccos(1 - 2 * x)  # Shefer (9)
                dgdx = 2 / np.sin(g)  # Shefer (10)
                dxdeta = -2 * self.m / eta**3  # next few lines are Shefer (14ish)
                dXdeta = self.dXdg(g) * dgdx * dxdeta
                Geta = eta - 1 - (self.ell + x) * self.X(g)
                dGdeta = 1 - self.X(g) * dxdeta - (self.ell + x) * dXdeta
                eta -= Geta / dGdeta
                niter += 1
            x = self.m / eta**2 - self.ell  # Shefer (9)
        elif self.cos2f < 0:
            x = 0.5
            Fx = 1
            niter = 0
            while np.abs(Fx) > self.eps and niter < self.maxiter:
                if 1 - 2 * x > 1 or 1 - 2 * x < -1:
                    raise RuntimeError("Invalid x")
                g = np.arccos(1 - 2 * x)  # Shefer (9)
                dgdx = 2 / np.sin(g)  # Shefer (10)
                eta = 1 + (self.ell + x) * self.X(g)  # Shefer (15)
                detadx = (self.ell + x) * self.dXdg(g) * dgdx + self.X(g)
                Fx = x - self.m / eta**2 + self.ell  # Shefer (15ish)
                dFdx = 1 + (2 * self.m / eta**3) * detadx
                x -= Fx / dFdx
                niter += 1
            eta = 1 + (self.ell + x) * self.X(np.arccos(1 - 2 * x))
        else:
            raise "Invalid value of cos2f"

        # Shefer (2)
        p = (0.5 * eta * self.kappa * self.sigma / self.tau)**2 / self.mu
        return p


class SheferTwoPosOrbitSolver(TwoPosOrbitSolver):
    """
    A class for solving two-position orbit determination problems using 
    Shefer's method. This class extends the `TwoPosOrbitSolver` base class 
    and implements a robust algorithm for determining the orbital parameter 
    `p` (semi-latus rectum) and auxiliary values based on Shefer's equations.

    Attributes:
        robust (bool): If True, enables a robust solution method that examines 
            all possible solutions when the initial guess fails.
        nExam (int): Number of points to examine when searching for zeros of 
            the function F(x) during the robust solution process.
        eps (float): Convergence tolerance for iterative calculations.
        maxiter (int): Maximum number of iterations allowed for convergence.
        m (float): A parameter related to the orbit determination problem.
        ell (float): A parameter related to the orbit determination problem.
        kappa (float): A constant used in orbital calculations.
        sigma (float): A constant used in orbital calculations.
        tau (float): A constant used in orbital calculations.
        mu (float): Standard gravitational parameter.
        lam (float): A parameter related to Shefer's equations.
        rbar (float): A normalized radial distance used in Shefer's calculations.

    Methods:
        alpha(x):
            Compute the alpha(x) function and its derivative based on Shefer's 
            equation (18).

        beta(xi):
            Compute the beta(xi) function and its derivative based on Shefer's 
            equations (A.4) and (A.9).

        Y(x):
            Compute the Y(x) function and its derivative based on Shefer's 
            equation (17).

        Yxi(xi):
            Compute the Y(xi) function using Shefer's equation (A.15).

        X(x):
            Compute the X(x) function and its derivative for elliptical orbits 
            (Shefer 19) and hyperbolic orbits (Shefer 20). Handles special 
            cases for small values of x.

        Z(xi):
            Compute the Z(xi) function and its derivative based on Shefer's 
            equation (A.5).

        D(x):
            Compute the D(x) function and its derivative based on Shefer's 
            equation (43).

        semiMajorAxis(x):
            Compute the semi-major axis a(x) and its derivative based on 
            Shefer's equation (42).

        Flam0(x):
            Compute Flam0(x) and its derivative based on Shefer's equations 
            (21) and (22).

        F(x):
            Compute F(x) and its derivative based on Shefer's equations (40) 
            and (41).

        G(xi):
            Compute G(xi) and its derivative based on Shefer's equations (A.7) 
            and (A.8).

        yPoly(y):
            Compute the polynomial for y and its derivative based on Shefer's 
            equations (44) and (38).

        _getInitialXGuess():
            Compute the initial guess for x using Shefer's step B. Handles 
            both cases where lam is zero and non-zero.

        _getInitialXiGuess():
            Compute the initial guess for xi using Shefer's equation (A.16).

        _getP():
            Compute the semi-latus rectum (p) using Shefer's algorithm. This 
            is the main result of the orbit determination process.

        _getAllP():
            Compute all possible values of p by finding zeros of the function 
            F(x) within a given range.

        _getEta(p):
            Compute the auxiliary value eta defined in Shefer's equation (2).

        solve():
            Solve the orbit determination problem. First attempts to find a 
            solution using Shefer's initial guess. If the solution fails, 
            employs a robust method to examine all possible zeros of F(x) 
            and determine a valid orbit.
    """
    def __init__(self, *args, **kwargs):
        self.robust = kwargs.pop('robust', False)
        self.nExam = kwargs.pop('nExam', 100)
        TwoPosOrbitSolver.__init__(self, *args, **kwargs)

    def alpha(self, x):
        """Evaluate alpha(x) from Shefer (18)."""
        return (self.rbar + self.kappa * (x - 0.5)), self.kappa

    def beta(self, xi):
        """Evaluate beta(xi) and its derivative from Shefer (A.4) and (A.9)."""
        val = self.rbar - 0.5 * self.kappa + xi * (self.rbar + 0.5 * self.kappa)
        grad = self.rbar + 0.5 * self.kappa
        return val, grad

    def Y(self, x):
        """Evaluate Y(x) and dY(x)/dx from Shefer (17)."""
        XVal, XGrad = self.X(x)
        alphaVal, alphaGrad = self.alpha(x)
        val = self.kappa + alphaVal * XVal
        # Evaluate dY(x)/dx using the chain rule
        grad = alphaVal * XGrad + alphaGrad * XVal
        return val, grad

    def Yxi(self, xi):
        betaVal, _ = self.beta(xi)
        return self.kappa + betaVal * self.Z(xi)[0]

    @staticmethod
    def X(x):
        """Evaluate X(x) function from Shefer (19) for elliptical orbits
        and (20) for hyperbolic orbits. The derivative of X(x) is given in
        (23).
        """
        import astropy.units as u
        if isinstance(x, u.Quantity):
            x = x.value
        if x < -1e-15:
            om2x = 1 - 2 * x
            xxm1 = x * (x - 1)
            val = om2x * np.sqrt(xxm1) - np.arcsinh(np.sqrt(-x))
            val /= 2 * (xxm1)**1.5
            grad = 4 - 3 * (om2x) * val
            grad /= -2 * xxm1
            return val, grad
        elif -1e-15 < x < 1e-15:
            grad = 8 / 5 + 64 / 35 * x
            val = grad * x + 4 / 3
            return val, grad
        elif x < 1.:
            om2x = 1 - 2 * x
            x1mx = x * (1 - x)
            val = np.arcsin(np.sqrt(x)) - om2x * np.sqrt(x1mx)
            val /= 2 * x1mx**1.5
            grad = 4 - 3 * om2x * val
            grad /= 2 * x1mx
            return val, grad
        else:  # x > 1.
            return np.inf  # X(x) actually asymptotes to Infinity as x -> 1

    @staticmethod
    def Z(xi):
        """Compute Z(xi) function from Shefer (A.5)."""
        num = (1 + xi)**2 * np.arctanh(np.sqrt(-xi)) - (1 - xi) * np.sqrt(-xi)
        denom = 2 * xi * np.sqrt(-xi)
        val = num / denom
        grad = (4 - (3 - xi) * val) / (2 * xi * (1 + xi))
        return val, grad

    def D(self, x):
        """Compute D(x) and its derivative, dD(x)/dx, from Shefer (43)."""
        aval, agrad = self.semiMajorAxis(x)
        val = self.tau * np.sqrt(self.mu) - 2 * np.pi * self.lam * aval**1.5
        grad = -3 * np.pi * self.lam * np.sqrt(aval) * agrad
        # print("----- D: ", x, aval, agrad, self.tau, self.mu, self.lam)
        return val, grad

    def semiMajorAxis(self, x):
        """Compute semi-major axis a(x) and its derivative, da(x)/dx, from
        Shefer (42).
        """
        assert x >= 0. and x <= 1., \
            "ERROR: invalid x in Shefer semi_major_axis: {}".format(x)
        alphaVal, alphaGrad = self.alpha(x)
        x1mx = x * (1 - x)
        val = alphaVal / (4 * x1mx)
        grad = (self.kappa * x1mx - alphaVal * (1 - 2 * x)) / (4 * x1mx**2)
        return val, grad

    def Flam0(self, x):  # Shefer (21), (22)
        Y, _ = self.Y(x)
        Ysq = Y**2
        alphaVal, _ = self.alpha(x)
        XVal, XGrad = self.X(x)
        val = alphaVal * Ysq - self.mu * self.tau**2
        grad = self.kappa * Ysq + 2 * alphaVal * Y * (self.kappa * XVal + alphaVal * XGrad)
        return val, grad

    def F(self, x):
        """Compute F(x) and dF(x)/dx from Shefer (40) and (41)."""
        if x >= 1.:
            # Should not get here
            val = 1.e16
            grad = 1.e16
        else:
            alphaVal, alphaGrad = self.alpha(x)
            YVal, YGrad = self.Y(x)
            DVal, DGrad = self.D(x)

            val = alphaVal * YVal**2 - DVal**2
            grad = alphaGrad * YVal**2 + alphaVal * 2 * YVal * YGrad - 2 * DVal * DGrad
        return val, grad

    def G(self, xi):
        """Compute G(xi) and its derivative from Shefer (A.7) and (A.8)."""
        betaVal, betaGrad = self.beta(xi)
        Y = self.Yxi(xi)
        Ysq = Y**2
        val = betaVal * Ysq - (1 + xi) * self.mu * self.tau**2
        ZVal, ZGrad = self.Z(xi)
        grad = (betaGrad * Ysq
                + 2 * betaVal * Y * (betaGrad * ZVal + betaVal * ZGrad)
                - self.mu * self.tau**2)
        return val, grad

    def yPoly(self, y):  # Shefer (44) and derivative
        coef = 0.6 * (self.rbar + 1.5 * self.kappa)
        ysqr = y * y
        den = ysqr + self.kappa
        x = (ysqr - self.rbar + 0.5 * self.kappa) / den  # Shefer (36)
        if x > 0 and x < 1 and self.lam > 0:
            dxdy = (self.kappa + 2 * self.rbar) * y / den**2
            Dval, Dgrad = self.D(x)
            val = (ysqr + coef) * y - 1.2 * Dval
            grad = 3 * ysqr + coef - 1.2 * Dgrad * dxdy
        else:
            val = (ysqr + coef) * y - 1.2 * self.tau * np.sqrt(self.mu)  # Shefer (38)
            grad = 3 * ysqr + coef
        return val, grad

    def _getInitialXGuess(self):
        # Shefer step B)
        if self.lam == 0:
            # Shefer (38)
            poly = [1.0, 0.0, 0.6 * (self.rbar + 1.5 * self.kappa),
                    -1.2 * (self.tau * np.sqrt(self.mu))]
            roots = np.roots(poly)
            # There should be exactly one positive real root.
            w = np.where((np.abs(roots.imag) < 3e-16) & (roots.real > 0))
            if len(w) > 1:
                raise RuntimeError(
                    "Found more than one positive, real root!  {}".format(roots)
                )
            if len(w) == 0:
                raise RuntimeError(
                    "Found no positive real roots!  {}".format(roots))
            y = roots[w][0].real
        else:
            y = np.sqrt(2 * self.rbar - self.kappa)
            y = newton_raphson(
                y, self.yPoly, eps=self.eps, maxiter=self.maxiter
            )
        ysqr = y * y
        # Shefer (36)
        x = (ysqr - self.rbar + 0.5 * self.kappa) / (ysqr + self.kappa)
        return x

    def _getInitialXiGuess(self):
        # Shefer (A.16)
        poly = [1.0, 0.0, (3. / 13.) * (self.rbar + 3.5 * self.kappa),
                -(12. / 13.) * (self.tau * np.sqrt(self.mu))]
        roots = np.roots(poly)
        # There should be exactly one positive real root.
        w = np.where((np.abs(roots.imag) < 3e-16) & (roots.real > 0))
        if len(w) > 1:
            raise RuntimeError(
                "Found more than one positive, real root!  {}".format(roots))
        if len(w) == 0:
            raise RuntimeError(
                "Found no positive real roots!  {}".format(roots))
        y = roots[w][0].real
        ysqr = y * y
        # Shefer (A.13)
        xi = (ysqr - self.rbar + 0.5 * self.kappa) / (self.rbar + 0.5 * self.kappa)
        return xi

    def _getP(self):
        """Get the semi-latus rectum. This is the final result of the Shefer algorithm."""
        x = self._getInitialXGuess()
        if x < -1. or x > 1.:
            xi = self._getInitialXiGuess()
            # Shefer step C)
            xi = newton_raphson(xi, self.G, eps=self.eps, maxiter=self.maxiter)
            alpha = self.beta(xi)[0] / (1 + xi)
        else:
            if x < 0.:
                self.lam = 0
            # Shefer step C)
            if self.lam == 0:
                x = newton_raphson(
                    x, self.Flam0, eps=self.eps, maxiter=self.maxiter
                )
            else:
                x = newton_raphson(
                    x, self.F, eps=self.eps, maxiter=self.maxiter
                )
            alpha = self.alpha(x)[0]
        p = self.sigma**2 / (4 * alpha)
        return p

    def _getAllP(self):
        xmin = 1e-15
        xmax = 1.0
        xs = find_all_zeros(lambda x: self.F(x)[0], xmin, xmax, n=self.nExam)
        return [self.sigma**2 / (4 * self.alpha(x)[0]) for x in xs]

    def _getEta(self, p):
        """Compute auxiliary value defined in Shefer (2)."""
        return 2 * np.sqrt(p * self.mu) * self.tau / (self.kappa * self.sigma)

    def solve(self):
        # Try normal solution first.
        orbit = TwoPosOrbitSolver.solve(self)
        if not self.robust:
            return orbit
        r1_test, _ = rv(orbit, self.t1)
        r2_test, _ = rv(orbit, self.t2)
        try:
            # 1 m tolerance...?
            # I'd really like better than this, but probably good enough for a
            # preliminary orbit determination.
            np.testing.assert_allclose(self.r1, r1_test, atol=1, rtol=0)
            np.testing.assert_allclose(self.r2, r2_test, atol=1, rtol=0)
        except AssertionError:
            pass
        else:
            # No AssertionError, so return early
            return orbit
        # Get here when there was an AssertionError above
        # Didn't find a solution with Shefer initial guess, so brute-force find
        # all the zeros of F and see if any of them yield a solution.
        for p in self._getAllP():
            orbit = self._finishOrbit(p)
            r1_test, _ = rv(orbit, self.t1)
            r2_test, _ = rv(orbit, self.t2)
            try:
                np.testing.assert_allclose(self.r1, r1_test, atol=1, rtol=0)
                np.testing.assert_allclose(self.r2, r2_test, atol=1, rtol=0)
            except AssertionError:
                pass
            else:
                return orbit
        # Get here when None of the p's work
        raise RuntimeError("Cannot find orbit")


class ThreeAngleOrbitSolver:
    """Determine orbit of satellite for set of three angle-only observations.

    Might only work well for smallish changes in sector position of
    observations.

    Parameters
    ----------
    e1, e2, e3 : (3,) array_like
        Unit vectors indicating observed directions (dimensionless).
    R1, R2, R3 : (3,) array_like
        Vectors indicating the positions of the observers (in meters).
    t1, t2, t3 : float or astropy.time.Time
        Times of observations.
        If float, then should correspond to GPS seconds; i.e., seconds since
        1980-01-06 00:00:00 UTC
    mu : float, optional
        Gravitational constant of central body in m^3/s^2.  (Default: Earth's
        gravitational constant in WGS84).
    tol : float, optional
        Tolerance to use to stop iteratively improving solution.
        (default 1e-8).
    maxiter : int, optional
        Maximum number of iterations to use.  (default: 100)
    """
    def __init__(self, e1, e2, e3, R1, R2, R3, t1, t2, t3, mu=EARTH_MU,
                 tol=1e-8, maxiter=100):
        # Follows Montenbruck and Gill section 2.4.2
        if isinstance(t1, Time):
            t1 = t1.gps
        if isinstance(t2, Time):
            t2 = t2.gps
        if isinstance(t3, Time):
            t3 = t3.gps
        self.e1 = e1
        self.e2 = e2
        self.e3 = e3
        self.t1 = t1
        self.t2 = t2
        self.t3 = t3
        self.R1 = R1
        self.R2 = R2
        self.R3 = R3
        self.mu = mu
        self.tol = tol
        self.maxiter = maxiter

    @lazy_property
    def d1(self):
        return np.cross(self.e2, self.e3)

    @lazy_property
    def d2(self):
        return np.cross(self.e3, self.e1)

    @lazy_property
    def d3(self):
        return np.cross(self.e1, self.e2)

    @lazy_property
    def D11(self):
        return np.dot(self.d1, self.R1)

    @lazy_property
    def D12(self):
        return np.dot(self.d1, self.R2)

    @lazy_property
    def D13(self):
        return np.dot(self.d1, self.R3)

    @lazy_property
    def D21(self):
        return np.dot(self.d2, self.R1)

    @lazy_property
    def D22(self):
        return np.dot(self.d2, self.R2)

    @lazy_property
    def D23(self):
        return np.dot(self.d2, self.R3)

    @lazy_property
    def D31(self):
        return np.dot(self.d3, self.R1)

    @lazy_property
    def D32(self):
        return np.dot(self.d3, self.R2)

    @lazy_property
    def D33(self):
        return np.dot(self.d3, self.R3)

    @lazy_property
    def D(self):
        return np.dot(self.e1, self.d1)

    @lazy_property
    def t21(self):
        return self.t2 - self.t1

    @lazy_property
    def t31(self):
        return self.t3 - self.t1

    @lazy_property
    def t32(self):
        return self.t3 - self.t2

    @lazy_property
    def t3231(self):
        return self.t32 / self.t31

    @lazy_property
    def t2131(self):
        return self.t21 / self.t31

    def _getRho(self, n1, n3):
        """Compute three unknown station-satellite distances from Montenbruck and Gill (2.129).
        """
        rho1 = -1 / (n1 * self.D) * (n1 * self.D11 - self.D12 + n3 * self.D13)
        rho2 = 1 / self.D * (n1 * self.D21 - self.D22 + n3 * self.D23)
        rho3 = -1 / (n3 * self.D) * (n1 * self.D31 - self.D32 + n3 * self.D33)
        return rho1, rho2, rho3

    def _getR(self, rho1, rho2, rho3):
        """Compute three Earth-satellite position vectors from Montenbruck and Gill (2.122).
        """
        r1 = self.R1 + rho1 * self.e1
        r2 = self.R2 + rho2 * self.e2
        r3 = self.R3 + rho3 * self.e3
        return r1, r2, r3

    def _getN(self, eta21, eta23):
        """Compute n1 and n3 terms from Montenbruck and Gill (2.132)."""
        n1 = eta21 * self.t3231
        n3 = eta23 * self.t2131
        return n1, n3

    def _getEta(self, r1, r2, t1, t2):
        """Use Shefer algorithm to improve eta estimates."""
        solver = SheferTwoPosOrbitSolver(r1, r2, t1, t2)
        p = solver._getP()
        eta = solver._getEta(p)
        return eta

    def solve(self):
        # initial estimate
        eta21 = eta23 = 1.0
        n1, n3 = self._getN(eta21, eta23)
        # iterate to improve eta21, eta23
        dn1 = np.inf
        dn3 = np.inf
        niter = 0
        while ((np.abs(dn1) > self.tol or np.abs(dn3) > self.tol)
               and niter < self.maxiter):
            rho1, rho2, rho3 = self._getRho(n1, n3)
            r1, r2, r3 = self._getR(rho1, rho2, rho3)
            eta1 = self._getEta(r2, r3, self.t2, self.t3)
            eta2 = self._getEta(r1, r3, self.t1, self.t3)
            eta3 = self._getEta(r1, r2, self.t1, self.t2)
            eta21 = eta2 / eta1
            eta23 = eta2 / eta3
            newn1, newn3 = self._getN(eta21, eta23)
            dn1, n1 = newn1 - n1, newn1
            dn3, n3 = newn3 - n3, newn3
            niter += 1
        solver = SheferTwoPosOrbitSolver(r1, r3, self.t1, self.t3)
        return solver.solve()



"""
Module to handle Earth satellite state vector manipulation and propagation.

Notes
-----
Although much of the Orbit class here is agnostic with respect to the
coordinate reference frame, we strongly recommend that positions and
velocities be specified in the GCRF frame.  Some classes in this module *do*
insist on this frame, (e.g., EarthObserver).  We will try to indicate where
the GCRF frame is required, but make no guarantees that we don't miss any
such instances.
"""


import numpy as np
import astropy
from .utils import (
    normSq, normed, lazy_property as _lazy_property, unitAngle3, sunPos,
    _gpsToTT, _wrapToPi, teme_to_gcrf, gcrf_to_teme, iers_interp
)
from .constants import EARTH_MU
from .propagator import KeplerianPropagator as _KeplerianPropagator

try:
    import erfa
except ImportError:
    # Let this raise
    import astropy._erfa as erfa


# Conversion routines for anomalies
def _ellipticalEccentricToTrueAnomaly(E, e):
    """Compute true anomaly from eccentric anomaly for elliptical orbit.

    Parameters
    ----------
    E : array_like
        Eccentric anomaly in radians.
    e : float
        Eccentricity

    Returns
    -------
    array_like
        True anomaly in radians.
    """
    beta = e / (1 + np.sqrt((1 - e) * (1 + e)))
    return E + 2 * np.arctan(beta * np.sin(E) / (1 - beta * np.cos(E)))


def _ellipticalTrueToEccentricAnomaly(v, e):
    """Compute eccentric anomaly from true anomaly for elliptical orbit.

    Parameters
    ----------
    v : array_like
        True anomaly in radians.
    e : float
        Eccentricity

    Returns
    -------
    array_like
        Eccentric anomaly in radians.
    """
    beta = e / (1 + np.sqrt(1 - e * e))
    return v - 2 * np.arctan(beta * np.sin(v) / (1 + beta * np.cos(v)))


def _ellipticalTrueToEccentricAnomalyMany(v, e):
    beta = e / (1 + np.sqrt(1 - e * e))
    return v - 2 * np.arctan(beta[:, None] * np.sin(v) / (1 + beta[:, None] * np.cos(v)))


def _hyperbolicEccentricToTrueAnomaly(H, e):
    """Compute hyperbolic true anomaly from hyperbolic eccentric anomaly for
    hyperbolic orbit.

    Parameters
    ----------
    H : array_like
        Hyperbolic eccentric anomaly in radians.
    e : float
        Eccentricity

    Returns
    -------
    array_like
        Hyperbolic true anomaly in radians.
    """
    return 2. * np.arctan(np.sqrt((e + 1) / (e - 1)) * np.tanh(H / 2))


def _hyperbolicEccentricToTrueAnomalyMany(H, e):
    return 2. * np.arctan(np.sqrt((e + 1) / (e - 1))[:, None] * np.tanh(H / 2))


def _hyperbolicTrueToEccentricAnomaly(v, e):
    """Compute hyperbolic eccentric anomaly from hyperbolic true anomaly for
    hyperbolic orbit.

    Parameters
    ----------
    v : array_like
        Hyperbolic true anomaly in radians.
    e : float
        Eccentricity

    Returns
    -------
    array_like
        Hyperbolic eccentric anomaly in radians.
    """
    return np.arcsinh(np.sqrt(e * e - 1) * np.sin(v) / (1 + e * np.cos(v)))


def _hyperbolicTrueToEccentricAnomalyMany(v, e):
    return np.arcsinh(
        np.sqrt(e * e - 1)[:, None] * np.sin(v) / (1 + e * np.cos(v))[:, None]
    )


def _ellipticalEccentricToMeanAnomaly(E, e):
    """Compute mean anomaly from eccentric anomaly for elliptical orbit.

    Parameters
    ----------
        E : array_like
            Eccentric anomaly in radians.
        e : float
            Eccentricity

    Returns
    -------
    array_like
        Mean anomaly in radians.
    """
    return E - e * np.sin(E)


@np.vectorize
def _ellipticalMeanToEccentricAnomaly(M, e):
    """Compute eccentric anomaly from mean anomaly for elliptical orbit.

    Parameters
    ----------
    M : array_like
        Mean anomaly in radians.
    e : float
        Eccentricity

    Returns
    -------
    array_like
        Eccentric anomaly in radians.
    """
    k1 = 3 * np.pi + 2
    k2 = np.pi - 1
    k3 = 6 * np.pi - 1
    A = 3 * k2 * k2 / k1
    B = k3 * k3 / (6 * k1)

    MM = _wrapToPi(M)  # map to [-Pi, Pi]
    if np.abs(MM) < 1. / 6:
        E = MM + e * (np.cbrt(6 * MM) - MM)
    else:
        if MM < 0:
            w = MM + np.pi
            E = MM + e * (A * w / (B - w) - np.pi - MM)
        else:
            w = MM - np.pi
            E = MM + e * (np.pi - A * w / (B - w) - MM)
    e1 = 1 - e
    noCancellationRisk = (e1 + E * E / 6) >= 0.1
    # three iterations.  each with one Halley and one Newton-Raphson step
    for _ in range(3):
        fdd = e * np.sin(E)
        fddd = e * np.cos(E)
        if noCancellationRisk:
            f = (E - fdd) - MM
            fd = 1 - fddd
        else:
            f = _eMeSinE(E, e) - MM
            s = np.sin(0.5 * E)
            fd = e1 + 2 * e * s * s
        dee = f * fd / (0.5 * f * fdd - fd * fd)

        w = fd + 0.5 * dee * (fdd + dee * fddd / 3)
        fd += dee * (fdd + 0.5 * dee * fddd)
        E -= (f - dee * (fd - w)) / fd
    E += M - MM
    return E


@np.vectorize
def _eMeSinE(E, e):
    """Accurate computation of E - e sin(E).
    Use when E is close to 0 and e close to 1,
    i.e., near periapsis of almost parabolic orbits.

    Parameters
    ----------
    E : array_like
        Eccentric anomaly in radians.
    e : float
        Eccentricity

    Returns
    -------
    array_like
        E - e sin(E)
    """
    x = (1 - e) * np.sin(E)
    mE2 = -E * E
    term = E
    d = 0
    x0 = np.nan
    while x != x0:
        d += 2
        term *= mE2 / (d * (d + 1))
        x0 = x
        x = x - term
    return x


def _hyperbolicEccentricToMeanAnomaly(H, e):
    """Compute hyperbolic mean anomaly from hyperbolic eccentric anomaly for
    hyperbolic orbit.

    Parameters
    ----------
    H : array_like
        Hyperbolic eccentric anomaly in radians.
    e : float
        Eccentricity

    Returns
    -------
    array_like
        Hyperbolic mean anomaly in radians.
    """
    return e * np.sinh(H) - H


@np.vectorize
def _hyperbolicMeanToEccentricAnomaly(M, e):
    """Compute hyperbolic eccentric anomaly from hyperbolic mean anomaly for
    hyperbolic orbit.

    Parameters
    ----------
    M : array_like
        Hyperbolic mean anomaly in radians.
    e : float
        Eccentricity

    Returns
    -------
    array_like
        Hyperbolic eccentric anomaly in radians.
    """
    if e < 1.6:
        if (-np.pi < M and M < 0.) or M > np.pi:
            H = M - e
        else:
            H = M + e
    else:
        if (e < 3.6 and abs(M) > np.pi):
            H = M - np.copysign(e, M)
        else:
            H = M / (e - 1)
    iter = 0
    while iter < 50:
        f3 = e * np.cosh(H)
        f2 = e * np.sinh(H)
        f1 = f3 - 1
        f0 = f2 - H - M
        f12 = 2 * f1
        d = f0 / f12
        fdf = f1 - d * f2
        ds = f0 / fdf
        shift = f0 / (fdf + ds * ds * f3 / 6)
        H -= shift
        if abs(shift) < 1e-12:
            return H
        iter += 1
    raise RuntimeError(
        "Unable to compute hyperbolic eccentric anomaly for M={}, e={}"
        .format(M, e)
    )


# Conversion routines for longitudes
def _ellipticalEccentricToTrueLongitude(lE, ex, ey):
    """Compute true longitude from eccentric longitude for elliptical orbit.

    Parameters
    ----------
    lE : array_like
        Eccentric longitude in radians.
    ex, ey : float
        Eccentricity vector components.

    Returns
    -------
    array_like
        True longitude in radians.
    """
    epsilon = np.sqrt(1 - ex * ex - ey * ey)
    cosLE = np.cos(lE)
    sinLE = np.sin(lE)
    num = ex * sinLE - ey * cosLE
    den = epsilon + 1 - ex * cosLE - ey * sinLE
    return lE + 2 * np.arctan(num / den)


def _ellipticalTrueToEccentricLongitude(lv, ex, ey):
    """Compute eccentric longitude from true longitude for elliptical orbit.

    Parameters
    ----------
    lv : array_like
        True longitude in radians.
    ex, ey : float
        Eccentricity vector components.

    Returns
    -------
    array_like
        Eccentric longitude in radians.
    """
    epsilon = np.sqrt(1 - ex * ex - ey * ey)
    cosLv = np.cos(lv)
    sinLv = np.sin(lv)
    num = ey * cosLv - ex * sinLv
    den = epsilon + 1 + ex * cosLv + ey * sinLv
    return lv + 2 * np.arctan(num / den)


def _ellipticalTrueToEccentricLongitudeMany(lv, ex, ey):
    # lv (n, m)
    # ex, ey (n,)
    epsilon = np.sqrt(1 - ex * ex - ey * ey)
    cosLv = np.cos(lv)
    sinLv = np.sin(lv)
    num = ey[:, None] * cosLv - ex[:, None] * sinLv
    den = epsilon[:, None] + 1 + ex[:, None] * cosLv + ey[:, None] * sinLv
    return lv + 2 * np.arctan(num / den)


def _hyperbolicEccentricToTrueLongitude(lH, ex, ey):
    """Compute hyperbolic true longitude from hyperbolic eccentric longitude for
    hyperbolic orbit.

    Parameters
    ----------
    lH : array_like
        Hyperbolic eccentric longitude in radians.
    ex, ey : float
        Eccentricity vector components.

    Returns
    -------
    array_like
        Hyperbolic true longitude in radians.
    """
    # is there a way to do this directly w/o computing e, w?
    e = np.sqrt(ex * ex + ey * ey)
    w = np.arctan2(ey, ex)
    H = lH - w
    v = _hyperbolicEccentricToTrueAnomaly(H, e)
    return v + w


def _hyperbolicTrueToEccentricLongitude(lv, ex, ey):
    """Compute hyperbolic eccentric longitude from hyperbolic true longitude for
    hyperbolic orbit.

    Parameters
    ----------
    lv : array_like
        Hyperbolic true longitude in radians.
    ex, ey : float
        Eccentricity vector components.

    Returns
    -------
    array_like
        Hyperbolic eccentric longitude in radians.
    """
    e = np.sqrt(ex * ex + ey * ey)
    w = np.arctan2(ey, ex)
    v = lv - w
    H = _hyperbolicTrueToEccentricAnomaly(v, e)
    return H + w


def _hyperbolicTrueToEccentricLongitudeMany(lv, ex, ey):
    e = np.sqrt(ex * ex + ey * ey)
    w = np.arctan2(ey, ex)
    v = lv - w[:, None]
    H = _hyperbolicTrueToEccentricAnomaly(v, e[:, None])
    return H + w[:, None]


def _ellipticalEccentricToMeanLongitude(lE, ex, ey):
    """Compute mean longitude from eccentric longitude for elliptical orbit.

    Parameters
    ----------
    lE : array_like
        Eccentric longitude in radians.
    ex, ey : float
        Eccentricity vector components.

    Returns
    -------
    array_like
        Mean longitude in radians.
    """
    return lE - ex * np.sin(lE) + ey * np.cos(lE)


def _ellipticalMeanToEccentricLongitude(lM, ex, ey):
    """Compute eccentric longitude from mean longitude for elliptical orbit.

    Parameters
    ----------
    lM : array_like
        Mean longitude in radians.
    ex, ey : float
        Eccentricity vector components.

    Returns
    -------
    array_like
        Eccentric longitude in radians.
    """
    shift = 1.0
    lEmlM = 0.0
    lE = lM
    cosLE = np.cos(lE)
    sinLE = np.sin(lE)
    # previous version iterated until a given convergence criterion was met.
    # This version always iterates exactly 6 times.  I found this is ~50% faster
    # in typical use (as part of RVSampler) since the operations are all
    # vectorized.
    for iter in range(6):
        f2 = ex * sinLE - ey * cosLE
        f1 = 1 - ex * cosLE - ey * sinLE
        f0 = lEmlM - f2
        f12 = 2 * f1
        shift = f0 * f12 / (f1 * f12 - f0 * f2)
        lEmlM -= shift
        lE = lM + lEmlM
        cosLE = np.cos(lE)
        sinLE = np.sin(lE)
    return lE


# Specialization for parallelizing over orbits and times simultaneously
def _ellipticalMeanToEccentricLongitudeMany(lM, ex, ey):
    shift = 1.0
    lEmlM = 0.0
    lE = lM
    cosLE = np.cos(lE)
    sinLE = np.sin(lE)
    # Always iterate exactly 6 times.  See above.
    for iter in range(6):
        f2 = np.array(sinLE)
        f1 = np.array(cosLE)

        f2 *= ex[:, None]
        cosLE *= ey[:, None]
        f2 -= cosLE

        f1 *= -ex[:, None]
        sinLE *= ey[:, None]
        f1 -= sinLE
        f1 += 1
        # above is equivalent to commented lines below, but rewritten to avoid
        # temporary array allocation as much as reasonable:
        # f2 = ex[:,None]*sinLE - ey[:,None]*cosLE
        # f1 = 1 - ex[:,None]*cosLE - ey[:,None]*sinLE
        f0 = lEmlM - f2

        shift = np.array(f0)
        shift *= f1
        shift *= 2
        f1 *= f1
        f1 *= 2
        f0 *= f2
        f1 -= f0
        shift /= f1
        # above block is equivalent to below
        # shift = f0*f12/(f1*f12 - f0*f2)

        lEmlM -= shift
        lE = lM + lEmlM
        cosLE = np.cos(lE)
        sinLE = np.sin(lE)
    return lE


def _hyperbolicEccentricToMeanLongitude(lH, ex, ey):
    """Compute hyperbolic mean longitude from hyperbolic eccentric longitude for
    hyperbolic orbit.

    Parameters
    ----------
    lH : array_like
        Hyperbolic eccentric longitude in radians.
    ex, ey : float
        Eccentricity vector components.

    Returns
    -------
    array_like
        Hyperbolic mean longitude in radians.
    """
    e = np.sqrt(ex * ex + ey * ey)
    w = np.arctan2(ey, ex)
    H = lH - w
    M = _hyperbolicEccentricToMeanAnomaly(H, e)
    return M + w


def _hyperbolicMeanToEccentricLongitude(lM, ex, ey):
    """Compute hyperbolic eccentric longitude from hyperbolic mean longitude for
    hyperbolic orbit.

    Parameters
    ----------
    lM : array_like
        Hyperbolic mean longitude in radians.
    ex, ey : float
        Eccentricity vector components.

    Returns
    -------
    array_like
        Hyperbolic eccentric longitude in radians.
    """
    e = np.sqrt(ex * ex + ey * ey)
    w = np.arctan2(ey, ex)
    M = lM - w
    H = _hyperbolicMeanToEccentricAnomaly(M, e)
    return H + w


def _hyperbolicMeanToEccentricLongitudeMany(lM, ex, ey):
    e = np.sqrt(ex * ex + ey * ey)
    w = np.arctan2(ey, ex)
    M = lM - w[:, None]
    H = _hyperbolicMeanToEccentricAnomaly(M, e[:, None])
    return H + w[:, None]


class Orbit:
    """
    Orbital state of one or more objects.

    This class represents one or more instantaneous (osculating) states of
    orbiting bodies.  Representations in Cartesian, Keplerian, and equinoctial
    elements are available.  The default initializer requires Cartesian
    elements, though class methods are also available to initialize with
    Keplerian or equinoctial elements.

    Note that generally this class can represent either a single scalar orbit,
    in which case attributes will generally be scalar floats, or a vector of
    orbits, in which case attributes will be ndarrays of floats.  For attributes
    that are intrinsically vectors even for a single orbit (position, velocity,
    ...), a 2d array will be used for a "vector of orbits", with the first
    dimension representing the different orbits.

    For simplicity, we will only indicate the "single scalar Orbit" dimensions
    in docstrings.  For a vector-Orbit, most scalar input arguments can be
    supplied as broadcastable arrays, and scalar attributes become array
    attributes.  When a multi-dimensional array is required, the first dimension
    is the one over different orbits.

    Parameters
    ----------
    r : (3,) array_like
        Position of orbiting object in meters.
    v : (3,) array_like
        Velocity of orbiting objects in meters per second.
    t : float or astropy.time.Time
        If float or array of float, then should correspond to GPS seconds; i.e.,
        seconds since 1980-01-06 00:00:00 UTC
    mu : float, optional
        Gravitational constant of central body in m^3/s^2.  (Default: Earth's
        gravitational constant in WGS84).

    Attributes
    ----------
    r : (3,) array_like
        Position of orbiting object in meters.
    v : (3,) array_like
        Velocity of orbiting object in meters per second.
    t : float
        GPS seconds; i.e., seconds since 1980-01-06 00:00:00 UTC
    mu : float
        Gravitational constant of central body in m^3/s^2
    a : float
        Semimajor axis in meters.
    hx, hy : float
        Components of the equinoctial inclination vector.
    ex, ey : float
        Components of the equinoctial eccentricity vector.
    lv : float
        True longitude in radians.
    lE : float
        Eccentric longitude in radians.
    lM : float
        Mean longitude in radians.
    e : float
        Keplerian eccentricity.
    i : float
        Keplerian inclination in radians.
    pa : float
        Keplerian periapsis argument in radians.
    raan : float
        Keplerian right ascension of the ascending node in radians.
    trueAnomaly : float
        Keplerian true anomaly in radians.
    eccentricAnomaly : float
        Keplerian eccentric anomaly in radians.
    meanAnomaly : float
        Keplerian mean anomaly in radians.
    period : float
        Orbital period in seconds.
    meanMotion : float
        Keplerian mean motion in radians per second.
    p : float
        Semi-latus rectum in meters.
    angularMomentum : (3,) array_like
        (Specific) angular momentum in m^2/s.
    energy: float
        (Specific) orbital energy in m^2/s^2.
    LRL : (3,) array_like
        Laplace-Runge-Lenz vector in m^3/s^2.
    periapsis : (3,) array_like
        Periapsis coordinate of orbit in meters.
    apoapsis : (3,) array_like
        Apoapsis coordinate of orbit in meters.
    equinoctialElements
    keplerianElements

    Methods
    -------
    fromKeplerianElements(a, e, i, pa, raan, trueAnomaly, t, mu)
        Construct an `Orbit` from Keplerian elements.
    fromEquinoctialElements(a, hx, hy, ex, ey, lv, t, mu)
        Construct an `Orbit` from Equinoctial elements.
    at(t, propagator)
        Propagate orbit to time t and return new Orbit.

    Notes
    -----
    Although much of the Orbit class here is agnostic with respect to the
    coordinate reference frame, we strongly recommend that positions and
    velocities be specified in the GCRF frame.  Some other classes in this
    module *do* insist on this frame, (e.g., EarthObserver).  We will try to
    indicate where the GCRF frame is required, but make no guarantees that we
    don't miss any such instances.
    """
    # Internally use equinoctial elements, calculating cartesian and keplerian
    # on demand.  Main ctor uses cartesian though
    def __init__(self, r, v, t, mu=EARTH_MU, propkw=None):
        if isinstance(t, astropy.time.Time):
            t = t.gps
        self.r, self.v = np.broadcast_arrays(r, v)
        self.propkw = dict() if propkw is None else propkw
        if self.r.ndim == 2:
            t = np.broadcast_to(t, self.r.shape[0])
            for key, value in self.propkw.items():
                self.propkw[key] = np.broadcast_to(value, self.r.shape[0])
        self.t = t
        self.mu = mu

    def __len__(self):
        if self.r.ndim == 1:
            return 1
        else:
            return self.r.shape[0]

    def __iter__(self):
        self._iter = 0
        return self

    def __next__(self):
        if self._iter < len(self):
            # propagate kozai directly if possible
            if 'kozaiMeanKeplerianElements' in self.__dict__:
                kMKE = self.__dict__['kozaiMeanKeplerianElements']
            else:
                kMKE = None
            out = Orbit(
                self.r[self._iter],
                self.v[self._iter],
                self.t[self._iter],
                propkw={k: v[self._iter] for k, v in self.propkw.items()}
            )
            if kMKE is not None:
                out.kozaiMeanKeplerianElements = kMKE
            # TODO: iterate through already-instantiated lazy_properties and
            # copy them over too.
            self._iter += 1
            return out
        else:
            raise StopIteration

    def __getitem__(self, idx):
        if 'kozaiMeanKeplerianElements' in self.__dict__:
            kMKE = self.__dict__['kozaiMeanKeplerianElements']
        else:
            kMKE = None
        out = Orbit(self.r[idx], self.v[idx], self.t[idx],
                    propkw={k: v[idx] for k, v in self.propkw.items()})
        if kMKE is not None:
            out.kozaiMeanKeplerianElements = kMKE
        return out

    def __hash__(self):
        if not hasattr(self, '_hash'):
            self.r.flags.writeable = False
            self.v.flags.writeable = False
            if self.r.ndim == 1:
                self._hash = hash((
                    self.r.data.tobytes(),
                    self.v.data.tobytes(),
                    self.t,
                    self.mu) + tuple([(k, v.data.tobytes()) for k, v in self.propkw.items()])
                )
            else:
                self.t.flags.writeable = False
                self._hash = hash((
                    self.r.data.tobytes(),
                    self.v.data.tobytes(),
                    self.t.data.tobytes(),
                    self.mu) + tuple([(k, v.data.tobytes()) for k, v in self.propkw.items()])
                )
        return self._hash

    def __eq__(self, rhs):
        if self.r.ndim != rhs.r.ndim:
            return False
        else:
            return (
                np.all(self.r == rhs.r) and np.all(self.v == rhs.v) and np.all(self.mu == rhs.mu) and np.all(self.t == rhs.t) and np.all([np.all(self.propkw[k] == rhs.propkw[k]) for k in self.propkw]))

    @classmethod
    def fromEquinoctialElements(
        cls, a, hx, hy, ex, ey, lv, t, mu=EARTH_MU, propkw=None,
    ):
        """Construct an Orbit from equinoctial elements.

        Parameters
        ----------
        a : float
            Semimajor axis in meters.
        hx, hy : float
            Components of the equinoctial inclination vector.
        ex, ey : float
            Components of the equinoctial eccentricity vector.
        lv : float
            True longitude in radians.
        t : float or astropy.time.Time
            If float, then should correspond to GPS seconds; i.e., seconds since
            1980-01-06 00:00:00 UTC
        mu : float, optional
            Gravitational constant of central body in m^3/s^2.  (Default:
            Earth's gravitational constant in WGS84).

        Returns
        -------
        Orbit
            The Orbit with given parameters.
        """
        import numbers
        if isinstance(t, astropy.time.Time):
            t = t.gps
        if not all(isinstance(x, numbers.Real) for x in (a, hx, hy, ex, ey, lv, t)):
            a, hx, hy, ex, ey, lv, t = np.broadcast_arrays(
                a, hx, hy, ex, ey, lv, t
            )
            scalar = False
        else:
            scalar = True

        obj = cls.__new__(cls)
        obj.a = a
        obj.hx = hx
        obj.hy = hy
        obj.ex = ex
        obj.ey = ey
        obj.lv = lv
        obj.t = t
        obj.mu = mu

        # set reference axes of orbital plane
        cls._setPQEquinoctial(obj)

        # set position and velocity
        if scalar:
            r, v = cls._rvFromEquinoctial(obj, lv=np.atleast_2d(lv))
            obj.r = r[0, 0]
            obj.v = v[0, 0]
            obj.propkw = propkw if propkw is not None else dict()
        else:
            r, v = cls._rvFromEquinoctial(obj, lv=lv[:, None])
            obj.r = r[:, 0]
            obj.v = v[:, 0]
            obj.propkw = propkw if propkw is not None else dict()
            for k, v in obj.propkw.items():
                obj.propkw[k] = np.broadcast_to(v, obj.a.shape[0])
        return obj

    @classmethod
    def fromKeplerianElements(
        cls, a, e, i, pa, raan, trueAnomaly, t, mu=EARTH_MU, propkw=None,
    ):
        """Construct an Orbit from Keplerian elements.

        Parameters
        ----------
        a : float
            Semimajor axis in meters.
        e : float
            Keplerian eccentricity.
        i : float
            Keplerian inclination in radians.
        pa : float
            Keplerian periapsis argument in radians.
        raan : float
            Keplerian right ascension of the ascending node in radians.
        trueAnomaly : float
            Keplerian true anomaly in radians.
        t : float or astropy.time.Time
            If float, then should correspond to GPS seconds; i.e., seconds since
            1980-01-06 00:00:00 UTC
        mu : float, optional
            Gravitational constant of central body in m^3/s^2.  (Default:
            Earth's gravitational constant in WGS84).

        Returns
        -------
        Orbit
            The Orbit with given parameters.
        """
        import numbers
        if isinstance(t, astropy.time.Time):
            t = t.gps
        if not all(isinstance(x, numbers.Real) for x in (a, e, i, pa, raan, trueAnomaly, t)):
            a, e, i, pa, raan, trueAnomaly, t = np.broadcast_arrays(
                a, e, i, pa, raan, trueAnomaly, t
            )
            scalar = False
        else:
            scalar = True

        obj = cls.__new__(cls)
        obj.a = a
        obj.e = e
        obj.i = i
        obj.pa = pa
        obj.raan = raan
        obj.trueAnomaly = trueAnomaly
        obj.t = t
        obj.mu = mu

        # set reference axes of orbital plane
        cls._setPQKeplerian(obj)

        # set position and velocity
        if scalar:
            r, v = cls._rvFromKeplerian(obj, np.atleast_2d(trueAnomaly))
            obj.r = r[0, 0]
            obj.v = v[0, 0]
            obj.propkw = propkw if propkw is not None else dict()
        else:
            r, v = cls._rvFromKeplerian(obj, trueAnomaly[:, None])
            obj.r = r[:, 0]
            obj.v = v[:, 0]
            obj.propkw = propkw if propkw is not None else dict()
            for k, v in obj.propkw.items():
                obj.propkw[k] = np.broadcast_to(v, obj.a.shape[0])
        return obj

    @classmethod
    def fromKozaiMeanKeplerianElements(
        cls, a, e, i, pa, raan, trueAnomaly, t, mu=EARTH_MU,
        _useTEME=False, propkw=None
    ):
        """Construct an Orbit from Kozai mean Keplerian elements.  By default,
        this method also converts from the TEME reference frame to the GCRF
        frame.  This is mainly to support TLEs and SGP4, so the default
        gravitational parameter is WGS84.

        Parameters
        ----------
        a : float
            Semimajor axis in meters.
        e : float
            Keplerian eccentricity.
        i : float
            Keplerian inclination in radians.
        pa : float
            Keplerian periapsis argument in radians.
        raan : float
            Keplerian right ascension of the ascending node in radians.
        trueAnomaly : float
            Keplerian true anomaly in radians.
        t : float or astropy.time.Time
            If float, then should correspond to GPS seconds; i.e., seconds since
            1980-01-06 00:00:00 UTC
        mu : float, optional
            Gravitational constant of central body in m^3/s^2.  (Default:
            Earth's gravitational constant in WGS84).

        Returns
        -------
        Orbit
            The Orbit with given parameters.
        """
        # _useTEME hidden bool kwarg indicates to keep SGP4 result in TEME
        # frame, instead of transforming to GCRF.  The default of False is
        # almost always appropriate.  The exception is when fitting for kozai
        # elements where it's faster to just work in the TEME frame.
        from sgp4.api import Satrec, WGS84

        if e <= 1:
            meanAnomaly = _ellipticalEccentricToMeanAnomaly(
                _ellipticalTrueToEccentricAnomaly(
                    trueAnomaly % (2 * np.pi), e
                ),
                e
            )
        else:
            meanAnomaly = _hyperbolicEccentricToMeanAnomaly(
                _hyperbolicTrueToEccentricAnomaly(
                    trueAnomaly % (2 * np.pi), e
                ),
                e
            )
        meanMotion = np.sqrt(mu / np.abs(a**3)) * 60.0  # rad/min

        if not isinstance(t, astropy.time.Time):
            t = astropy.time.Time(t, format='gps')
        mjd = t.mjd
        # epoch = mjd - astropy.time.Time("1949-12-31T00:00:00").mjd
        epoch = mjd - 33281.0

        sat = Satrec()
        sat.sgp4init(
            WGS84,           # gravity model
            'i',             # 'a' = old AFSPC mode, 'i' = improved mode
            0,               # satnum: Satellite number
            epoch,           # epoch: days since 1949 December 31 00:00 UT
            0.0,             # bstar: drag coefficient (kg/m2er)
            0.0,             # ndot: ballistic coefficient (revs/day)
            0.0,             # nddot: second derivative of mean motion (revs/day^3)
            e,               # ecco: eccentricity
            pa,              # argpo: argument of perigee (radians)
            i,               # inclo: inclination (radians)
            meanAnomaly,     # mo: mean anomaly (radians)
            meanMotion,      # no_kozai: mean motion (radians/minute)
            raan,            # nodeo: right ascension of ascending node (radians)
        )
        e, r, v = sat.sgp4_tsince(0)
        r = np.array(r)
        v = np.array(v)
        if not _useTEME:
            rot = teme_to_gcrf(t)
            r = rot @ r
            v = rot @ v
        r *= 1e3  # km -> m
        v *= 1e3  # km/s -> m/s
        return Orbit(r, v, t, mu=mu, propkw=propkw)

    @_lazy_property
    def kozaiMeanKeplerianElements(self):
        """Kozai mean Keplerian elements in TEME frame
        (a, e, i, pa, raan, trueAnomaly)
        """
        from scipy.optimize import least_squares
        # I don't know of a good closed form expression for Kozai elements, so
        # instead, just solve for them using SGP4 and a desired output state
        # vector.  The osculating elements are a good initial guess.

        rot = gcrf_to_teme(self.t)
        rTEME = rot @ self.r
        vTEME = rot @ self.v

        def resid(p):
            a, e, i, pa, raan, trueAnomaly = p
            orb = Orbit.fromKozaiMeanKeplerianElements(
                a, e, i, pa, raan, trueAnomaly,
                self.t,
                _useTEME=True  # stay in TEME frame
            )
            resid = np.hstack([(orb.r - rTEME) / 1e4, orb.v - vTEME])
            # print(f"{a:8.4f} {e:8.4f} {i:8.4f} {pa:8.4f} {raan:8.4f} {trueAnomaly:8.4f}", end='')
            # print(f"  {resid[0]:12.4f} {resid[1]:12.4f} {resid[2]:12.4f} {resid[3]:12.4f} {resid[4]:12.4f} {resid[5]:12.4f}")
            return resid
        lbounds = [0.0, 0.0, -np.inf, -np.inf, -np.inf, -np.inf]
        ubounds = [np.inf, np.inf, np.inf, np.inf, np.inf, np.inf]
        opt = least_squares(
            resid,
            np.array(self.keplerianElements),
            xtol=1e-15, ftol=None, gtol=None,
            bounds=[lbounds, ubounds]
        )
        # Sometimes we seem to get stuck in a local minimum, particularly when
        # the eccentricity is close to zero.  So check if we've reached adequate
        # convergence, and try a (not-random) restart if not.
        if np.any(np.abs(opt.fun) > 1e-3):
            opt = least_squares(
                resid,
                opt.x + np.ones(6) * 1e-3,
                xtol=1e-15, ftol=None, gtol=None,
                bounds=[lbounds, ubounds]
            )
            # One final attempt
            if np.any(np.abs(opt.fun) > 1e-3):
                opt = least_squares(
                    resid,
                    opt.x + np.array([-1, 1, -1, 1, -1, 1]) * 1e-3,
                    xtol=1e-15, ftol=None, gtol=None,
                    bounds=[lbounds, ubounds]
                )
        return opt.x

    @classmethod
    def fromTLE(cls, sat_name, tle_filename, propkw=None):
        """Construct an Orbit from a TLE file

        Parameters
        ----------
        sat_name : str
            NORAD name of the satellite
        tle_filename : str
            Path and name of file where TLE is

        Returns
        -------
        Orbit
        """
        from .io import read_tle
        tle = read_tle(sat_name, tle_filename)
        return cls.fromTLETuple(tle, propkw=propkw)

    @classmethod
    def fromTLETuple(cls, tle, propkw=None):
        """Construct an Orbit from a TLE tuple

        Parameters
        ----------
        tle : 2-tuple of str
            Line1 and Line2 of TLE as strings

        Returns
        -------
        Orbit
        """
        from .io import _rvt_from_tle_tuple, parse_tle
        # Should set Kozai elements here directly from TLE.
        a, e, i, pa, raan, trueAnomaly, t = parse_tle(tle)
        r, v, t2 = _rvt_from_tle_tuple(tle)
        assert abs(t - t2) < 1e-5  # 10 microsec
        # Transform TEME -> GCRF
        rot = teme_to_gcrf(t)
        r = rot @ r
        v = rot @ v
        obj = Orbit(r, v, t, mu=EARTH_MU)
        obj.kozaiMeanKeplerianElements = a, e, i, pa, raan, trueAnomaly
        obj.propkw = propkw if propkw is not None else dict()
        return obj

    def at(self, t, propagator=_KeplerianPropagator()):
        """Propagate this orbit to time t.

        Parameters
        ----------
        t : float or astropy.time.Time
            If float, then should correspond to GPS seconds; i.e., seconds since
            1980-01-06 00:00:00 UTC
        propagator : Propagator, optional
            The propagator instance to use.

        Returns
        -------
        Orbit
            Propagated Orbit.
        """
        from .compute import rv
        r, v = rv(self, t, propagator=propagator)
        return Orbit(r, v, t, mu=self.mu, propkw=self.propkw)

    def __repr__(self):
        out = "Orbit(r={!r}, v={!r}, t={!r}".format(self.r, self.v, self.t)
        if self.mu != EARTH_MU:
            out += ", mu={!r}".format(self.mu)
        out += ")"
        return out

    def _rvFromEquinoctial(self, lv=None, lE=None):
        # self is (n,) or scalar
        # assume lv/lE is always explicitly (n, m)
        # always return (n, m, 3)
        # calling code will squeeze
        # No checks done, but exactly one of lv, lE should be set.

        n = len(np.atleast_1d(self.a))
        if lv is not None:
            m = lv.shape[1]
        else:
            m = lE.shape[1]
        r = np.empty((n, m, 3))
        v = np.empty((n, m, 3))

        wBound = np.atleast_1d(self.a) > 0
        if np.any(wBound):
            a = np.atleast_1d(self.a)[wBound]
            ex = np.atleast_1d(self.ex)[wBound]
            ey = np.atleast_1d(self.ey)[wBound]
            ex2 = ex * ex
            ey2 = ey * ey
            e2 = ex2 + ey2
            exey = ex * ey
            eta = 1. + np.sqrt(1 - e2)
            beta = 1. / eta

            if lE is None:
                lEb = _ellipticalTrueToEccentricLongitudeMany(
                    lv[wBound], ex, ey
                )
            else:
                lEb = lE[wBound]

            cLe = np.cos(lEb)
            sLe = np.sin(lEb)
            exCeyS = ex[:, None] * cLe + ey[:, None] * sLe

            # avoid temporary arrays as much as possible.
            x = (1 - beta * ey2)[:, None] * cLe
            x += (beta * exey)[:, None] * sLe
            x -= ex[:, None]
            x *= a[:, None]
            y = (1 - beta * ex2)[:, None] * sLe
            y += (beta * exey)[:, None] * cLe
            y -= ey[:, None]
            y *= a[:, None]

            factor = np.sqrt(self.mu / a)[:, None] / (1. - exCeyS)
            xDot = factor * (-sLe + (beta * ey)[:, None] * exCeyS)
            yDot = factor * (cLe - (beta * ex)[:, None] * exCeyS)

            rb = np.einsum("ab,ac->abc", x, np.atleast_2d(self._pEq)[wBound])
            rb += np.einsum("ab,ac->abc", y, np.atleast_2d(self._qEq)[wBound])
            vb = np.einsum(
                "ab,ac->abc",
                xDot,
                np.atleast_2d(self._pEq)[wBound]
            )
            vb += np.einsum(
                "ab,ac->abc",
                yDot,
                np.atleast_2d(self._qEq)[wBound]
            )
            # checking if can avoid the array insert below.  The check is almost
            # free (speedwise), so usually a good idea.
            if np.all(wBound):
                return rb, vb
            r[wBound] = rb
            v[wBound] = vb

        if np.any(~wBound):
            # Copy Keplerian formula for the moment; optimize some other day...
            a = np.atleast_1d(self.a)[~wBound]
            ex = np.atleast_1d(self.ex)[~wBound]
            ey = np.atleast_1d(self.ey)[~wBound]
            e2 = ex * ex + ey * ey
            e = np.sqrt(e2)
            lonPa = np.arctan2(ey, ex)
            if lv is None:
                lvub = _hyperbolicTrueToEccentricLongitudeMany(
                    lE[~wBound], ex, ey
                )
            else:
                lvub = lv[~wBound]
            trueAnomaly = np.atleast_1d(lvub - lonPa[:, None])  # (n,m)

            sinV = np.sin(trueAnomaly)  # (n,m)
            cosV = np.cos(trueAnomaly)
            f = a * (1 - e * e)  # (n,)
            posFactor = f[:, None] / (1 + e[:, None] * cosV)  # (n,m)
            velFactor = np.sqrt(self.mu / f[:, None])  # (n,m)

            x = posFactor * cosV  # (n,m)
            y = posFactor * sinV
            xDot = -velFactor * sinV
            yDot = velFactor * (e[:, None] + cosV)

            # r = x*self._pK + y*self._qK
            # v = xDot*self._pK + yDot*self._qK

            s, c = np.sin(lonPa), np.cos(lonPa)  # (n,)
            R = np.array([[c, -s], [s, c]])  # (2, 2, n)
            rp = np.array([x, y])  # (2, n, m)
            rp = np.einsum("abc,bcd->acd", R, rp)
            rpDot = np.array([xDot, yDot])  # (2, n, m)
            rpDot = np.einsum("abc,bcd->acd", R, rpDot)  # (2, n, m)

            rub = np.einsum("nm,nt->nmt", rp[0], np.atleast_2d(self._pEq))
            rub += np.einsum("nm,nt->nmt", rp[1], np.atleast_2d(self._qEq))
            vub = np.einsum("nm,nt->nmt", rpDot[0], np.atleast_2d(self._pEq))
            vub += np.einsum("nm,nt->nmt", rpDot[1], np.atleast_2d(self._qEq))
            if np.all(~wBound):
                return rub, vub
            r[~wBound] = rub
            v[~wBound] = vub

        return r, v

    # Specialization for parallelizing over orbits
    @staticmethod
    def _rvFromEquinoctialMany(a, ex, ey, mu, pEq, qEq, lE):
        if np.any(a < 0):
            raise NotImplementedError(
                "Parallelization over hyperbolic orbits not implemented"
            )
        exey = ex * ey
        ex2 = ex * ex
        ey2 = ey * ey
        e2 = ex2 + ey2
        eta = 1. + np.sqrt(1 - e2)
        beta = 1. / eta

        cLe = np.cos(lE)
        sLe = np.sin(lE)
        exCeyS = ex[:, None] * cLe + ey[:, None] * sLe

        x = a[:, None] * (
            (1 - beta * ey2)[:, None] * cLe + (beta * exey)[:, None] * sLe - ex[:, None]
        )
        y = a[:, None] * (
            (1 - beta * ex2)[:, None] * sLe + (beta * exey)[:, None] * cLe - ey[:, None]
        )

        factor = np.sqrt(mu / a)[:, None] / (1. - exCeyS)
        xDot = factor * (-sLe + (beta * ey)[:, None] * exCeyS)
        yDot = factor * (cLe - (beta * ex)[:, None] * exCeyS)

        r = np.einsum("ab,ac->abc", x, pEq)
        r += np.einsum("ab,ac->abc", y, qEq)
        v = np.einsum("ab,ac->abc", xDot, pEq)
        v += np.einsum("ab,ac->abc", yDot, qEq)
        return r, v

    def _rvFromKeplerian(self, trueAnomaly):
        # self is (n,) or scalar
        # trueAnomaly is (n, m)
        # always return (n, m, 3); calling code will squeeze
        n = len(np.atleast_1d(self.a))
        m = trueAnomaly.shape[1]
        r = np.empty((n, m, 3))
        v = np.empty((n, m, 3))

        wBound = np.atleast_1d(self.a) > 0
        if np.any(wBound):
            a = np.atleast_1d(self.a)
            e = np.atleast_1d(self.e)[wBound]
            uME2 = (1 - e) * (1 + e)
            s1Me2 = np.sqrt(uME2)
            E = _ellipticalTrueToEccentricAnomalyMany(
                trueAnomaly[wBound],
                e
            )
            cosE = np.cos(E)
            sinE = np.sin(E)

            # coordinates of position and velocity in the orbital plane
            x = a[:, None] * (cosE - e[:, None])
            y = a[:, None] * sinE * s1Me2[:, None]
            factor = np.sqrt(self.mu / a[:, None]) / (1 - e[:, None] * cosE)
            xDot = -sinE * factor
            yDot = cosE * s1Me2[:, None] * factor
            rb = np.einsum("ab,ac->abc", x, np.atleast_2d(self._pK)[wBound])
            rb += np.einsum("ab,ac->abc", y, np.atleast_2d(self._qK)[wBound])
            vb = np.einsum("ab,ac->abc", xDot, np.atleast_2d(self._pK)[wBound])
            vb += np.einsum("ab,ac->abc", yDot, np.atleast_2d(self._qK)[wBound])
            r[wBound] = rb
            v[wBound] = vb
        if np.any(~wBound):
            sinV = np.atleast_1d(np.sin(trueAnomaly[~wBound]))
            cosV = np.atleast_1d(np.cos(trueAnomaly[~wBound]))
            a = np.atleast_1d(self.a)[~wBound]
            e = np.atleast_1d(self.e)[~wBound]
            f = a * (1 - e * e)
            posFactor = f[:, None] / (1 + e[:, None] * cosV)
            velFactor = np.sqrt(self.mu / f[:, None])

            x = posFactor * cosV
            y = posFactor * sinV
            xDot = -velFactor * sinV
            yDot = velFactor * (e[:, None] + cosV)

            rub = np.einsum("ab,ac->abc", x, np.atleast_2d(self._pK)[~wBound])
            rub += np.einsum("ab,ac->abc", y, np.atleast_2d(self._qK)[~wBound])
            vub = np.einsum("ab,ac->abc", xDot, np.atleast_2d(self._pK)[~wBound])
            vub += np.einsum("ab,ac->abc", yDot, np.atleast_2d(self._qK)[~wBound])
            r[~wBound] = rub
            v[~wBound] = vub

        return r, v

    def _setEquinoctial(self):
        # set equinoctial elements from state vectors
        # Compute semimajor axis
        r = np.atleast_2d(self.r)
        v = np.atleast_2d(self.v)
        r2 = normSq(r)
        rnorm = np.sqrt(r2)
        v2 = normSq(v)
        rv2OverMu = rnorm * v2 / self.mu

        # negative for hyperbolic
        a = rnorm / (2 - rv2OverMu)
        muA = a * self.mu

        # Compute inclination vector
        w = np.cross(r, v)
        wnormed = normed(w)
        d = 1. / (1 + wnormed[:, 2])
        hx = -d * wnormed[:, 1]
        hy = d * wnormed[:, 0]

        # Compute true longitude argument
        cLv = (r[:, 0] - d * r[:, 2] * wnormed[:, 0]) / rnorm
        sLv = (r[:, 1] - d * r[:, 2] * wnormed[:, 1]) / rnorm
        lv = np.arctan2(sLv, cLv)

        # Compute eccentricity vector
        # Separate maths for bound/unbound orbits
        e = np.empty_like(a)
        ex = np.empty_like(a)
        ey = np.empty_like(a)
        trueAnomaly = np.empty_like(a)
        lonPa = np.empty_like(a)

        wBound = a > 0

        # bound orbits first
        eS = np.einsum("ab,ab->a", r, v) / np.sqrt(np.abs(muA))
        eC = rv2OverMu - 1
        if np.any(wBound):
            eSE = eS[wBound]
            eCE = eC[wBound]
            e2 = eCE * eCE + eSE * eSE
            e[wBound] = np.sqrt(e2)
            f = eCE - e2
            g = np.sqrt(1 - e2) * eSE
            ex[wBound] = a[wBound] * (f * cLv[wBound] + g * sLv[wBound])
            ex[wBound] /= rnorm[wBound]
            ey[wBound] = a[wBound] * (f * sLv[wBound] - g * cLv[wBound])
            ey[wBound] /= rnorm[wBound]
            lonPa[wBound] = np.arctan2(ey[wBound], ex[wBound])
            trueAnomaly[wBound] = _wrapToPi(lv[wBound] - lonPa[wBound])

        # now unbound orbits
        if np.any(~wBound):
            e[~wBound] = np.sqrt(1 - normSq(w[~wBound]) / muA[~wBound])
            eSH = eS[~wBound]
            eCH = eC[~wBound]
            H = 0.5 * np.log((eCH + eSH) / (eCH - eSH))
            trueAnomaly[~wBound] = _wrapToPi(
                _hyperbolicEccentricToTrueAnomalyMany(
                    H[:, None], e[~wBound]
                )[:, 0]
            )
            lonPa[~wBound] = _wrapToPi(lv[~wBound] - trueAnomaly[~wBound])
            ex[~wBound] = e[~wBound] * np.cos(lonPa[~wBound])
            ey[~wBound] = e[~wBound] * np.sin(lonPa[~wBound])

        # now populate properties and squeeze as necessary.
        if self.r.ndim == 1:
            a = a[0]
            e = e[0]
            hx = hx[0]
            hy = hy[0]
            ex = ex[0]
            ey = ey[0]
            lv = lv[0]
            trueAnomaly = trueAnomaly[0]
            lonPa = lonPa[0]
        self.a = a
        self.e = e
        self.hx = hx
        self.hy = hy
        self.ex = ex
        self.ey = ey
        self.lv = lv
        self.trueAnomaly = trueAnomaly
        self.lonPa = lonPa

    def _setKeplerian(self):
        # set keplerian elements from state vectors
        # Let _setEquinoctial take care of setting a, e, trueAnomaly
        self._setEquinoctial()
        r = np.atleast_2d(self.r)
        v = np.atleast_2d(self.v)
        w = normed(np.cross(r, v))
        i = unitAngle3(w, np.array([0, 0, 1.]))
        wr = np.cross(np.array([0, 0, 1.]), w)
        raan = np.arctan2(wr[:, 1], wr[:, 0])
        pa = _wrapToPi(np.atleast_1d(self.lonPa) - raan)
        if self.r.ndim == 1:
            i = i[0]
            raan = raan[0]
            pa = pa[0]
        self.i = i
        self.raan = raan
        self.pa = pa

    def _setPQEquinoctial(self):
        hx2 = self.hx * self.hx
        hy2 = self.hy * self.hy
        hxhy = self.hx * self.hy
        factH = 1. / (1 + hx2 + hy2)

        # Reference axes defining orbital plane
        self._pEq = (np.array([1 + hx2 - hy2, 2 * hxhy, -2 * self.hy]) * factH).T
        self._qEq = (np.array([2 * hxhy, 1 - hx2 + hy2, 2 * self.hx]) * factH).T

    def _setPQKeplerian(self):
        cosRaan = np.cos(self.raan)
        sinRaan = np.sin(self.raan)
        cosPa = np.cos(self.pa)
        sinPa = np.sin(self.pa)
        cosI = np.cos(self.i)
        sinI = np.sin(self.i)
        crcp = cosRaan * cosPa
        crsp = cosRaan * sinPa
        srcp = sinRaan * cosPa
        srsp = sinRaan * sinPa
        self._pK = np.array([
            crcp - cosI * srsp, srcp + cosI * crsp, sinI * sinPa
        ]).T
        self._qK = np.array([
            -crsp - cosI * srcp, -srsp + cosI * crcp, sinI * cosPa
        ]).T

    @_lazy_property
    def _pEq(self):
        self._setPQEquinoctial()
        return self._pEq

    @_lazy_property
    def _qEq(self):
        self._setPQEquinoctial()
        return self._qEq

    @_lazy_property
    def _pK(self):
        self._setPQKeplerian()
        return self._pK

    @_lazy_property
    def _qK(self):
        self._setPQKeplerian()
        return self._qK

    @_lazy_property
    def a(self):
        """Semimajor axis in meters.
        """
        self._setEquinoctial()
        return self.a

    @_lazy_property
    def hx(self):
        """First component of equinoctial inclination vector.
        """
        self._setEquinoctial()
        return self.hx

    @_lazy_property
    def hy(self):
        """Second component of equinoctial inclination vector.
        """
        self._setEquinoctial()
        return self.hy

    @_lazy_property
    def ex(self):
        """First component of equinoctial eccentricity vector.
        """
        self._setEquinoctial()
        return self.ex

    @_lazy_property
    def ey(self):
        """Second component of equinoctial eccentricity vector.
        """
        self._setEquinoctial()
        return self.ey

    @_lazy_property
    def lv(self):
        """True longitude in radians.
        """
        self._setEquinoctial()
        return self.lv

    @_lazy_property
    def lE(self):
        """Eccentric longitude in radians.
        """
        import numbers
        if isinstance(self.a, numbers.Real):
            if self.a > 0:
                return _ellipticalTrueToEccentricLongitude(
                    self.lv, self.ex, self.ey
                )
            else:
                return _hyperbolicTrueToEccentricLongitude(
                    self.lv, self.ex, self.ey
                )
        else:
            wBound = self.a > 0
            lE = np.empty_like(self.lv)
            if np.any(wBound):
                lE[wBound] = _ellipticalTrueToEccentricLongitude(
                    self.lv[wBound], self.ex[wBound], self.ey[wBound]
                )
            if np.any(~wBound):
                lE[~wBound] = _hyperbolicTrueToEccentricLongitude(
                    self.lv[~wBound], self.ex[~wBound], self.ey[~wBound]
                )
            return lE

    @_lazy_property
    def lM(self):
        """Mean longitude in radians.
        """
        import numbers
        if isinstance(self.a, numbers.Real):
            if self.a > 0:
                return _ellipticalEccentricToMeanLongitude(
                    self.lE, self.ex, self.ey
                )
            else:
                return _hyperbolicEccentricToMeanLongitude(
                    self.lE, self.ex, self.ey
                )
        else:
            wBound = self.a > 0
            lM = np.empty_like(self.lv)
            if np.any(wBound):
                lM[wBound] = _ellipticalEccentricToMeanLongitude(
                    self.lE[wBound], self.ex[wBound], self.ey[wBound]
                )
            if np.any(~wBound):
                lM[~wBound] = _hyperbolicEccentricToMeanLongitude(
                    self.lE[~wBound], self.ex[~wBound], self.ey[~wBound]
                )
            return lM

    @_lazy_property
    def e(self):
        """Eccentricity.
        """
        return np.sqrt(self.ex * self.ex + self.ey * self.ey)

    @_lazy_property
    def i(self):
        """Inclination in radians.
        """
        self._setKeplerian()
        return self.i

    @_lazy_property
    def pa(self):
        """Periapsis argument in radians.
        """
        self._setKeplerian()
        return self.pa

    @_lazy_property
    def lonPa(self):
        """Longitude of periapsis argument in radians.
        """
        self._setEquinoctial()
        return self.lonPa

    @_lazy_property
    def raan(self):
        """Right ascension of the ascending node in radians.
        """
        self._setKeplerian()
        return self.raan

    @_lazy_property
    def trueAnomaly(self):
        """True anomaly in radians.
        """
        self._setEquinoctial()
        return self.trueAnomaly

    @_lazy_property
    def eccentricAnomaly(self):
        """Eccentric anomaly in radians.
        """
        import numbers
        if isinstance(self.a, numbers.Real):
            if self.a > 0:
                return _ellipticalTrueToEccentricAnomaly(
                    self.trueAnomaly, self.e
                )
            else:
                return _hyperbolicTrueToEccentricAnomaly(
                    self.trueAnomaly, self.e
                )
        else:
            wBound = self.a > 0
            eccentricAnomaly = np.empty_like(self.trueAnomaly)
            if np.any(wBound):
                eccentricAnomaly[wBound] = _ellipticalTrueToEccentricAnomaly(
                    self.trueAnomaly[wBound], self.e[wBound]
                )
            if np.any(~wBound):
                eccentricAnomaly[~wBound] = _hyperbolicTrueToEccentricAnomaly(
                    self.trueAnomaly[~wBound], self.e[~wBound]
                )
            return eccentricAnomaly

    @_lazy_property
    def meanAnomaly(self):
        """Mean anomaly in radians.
        """
        import numbers
        if isinstance(self.a, numbers.Real):
            if self.a > 0:
                return _ellipticalEccentricToMeanAnomaly(
                    _ellipticalTrueToEccentricAnomaly(self.trueAnomaly, self.e),
                    self.e
                )
            else:
                return _hyperbolicEccentricToMeanAnomaly(
                    _hyperbolicTrueToEccentricAnomaly(self.trueAnomaly, self.e),
                    self.e
                )
        else:
            wBound = self.a > 0
            meanAnomaly = np.empty_like(self.trueAnomaly)
            if np.any(wBound):
                meanAnomaly[wBound] = _ellipticalEccentricToMeanAnomaly(
                    _ellipticalTrueToEccentricAnomaly(
                        self.trueAnomaly[wBound], self.e[wBound]
                    ),
                    self.e[wBound]
                )
            if np.any(~wBound):
                meanAnomaly[~wBound] = _hyperbolicEccentricToMeanAnomaly(
                    _hyperbolicTrueToEccentricAnomaly(
                        self.trueAnomaly[~wBound], self.e[~wBound]
                    ),
                    self.e[~wBound]
                )
            return meanAnomaly

    @_lazy_property
    def period(self):
        """Orbital period in seconds.
        """
        import numbers
        if isinstance(self.a, numbers.Real):
            return 2 * np.pi / self.meanMotion if self.a > 0 else np.inf
        else:
            wBound = self.a > 0
            period = np.empty_like(self.a)
            if np.any(wBound):
                period[wBound] = 2 * np.pi / self.meanMotion[wBound]
            if np.any(~wBound):
                period[~wBound] = np.inf
            return period

    @_lazy_property
    def meanMotion(self):
        """Mean motion in radians per second.
        """
        return np.sqrt(self.mu / np.abs(self.a**3))

    @property
    def equinoctialElements(self):
        """Equinoctial elements (a, hx, hy, ex, ey, lv).
        """
        return self.a, self.hx, self.hy, self.ex, self.ey, self.lv

    @property
    def keplerianElements(self):
        """Keplerian elements (a, e, i, pa, raan, trueAnomaly).
        """
        return self.a, self.e, self.i, self.pa, self.raan, self.trueAnomaly

    @_lazy_property
    def p(self):
        """Semi-latus rectum in meters.
        """
        return self.a * (1.0 - self.e**2)

    @_lazy_property
    def angularMomentum(self):
        """(Specific) angular momentum vector in m^2/s.
        """
        return np.cross(self.r, self.v)

    @_lazy_property
    def energy(self):
        """(Specific) orbital energy in m^2/s^2.
        """
        return -0.5 * self.mu / self.a

    @_lazy_property
    def LRL(self):
        """Laplace-Runge-Lenz vector in m^3/s^2.
        """
        return -self.mu * normed(self.r) - np.cross(self.angularMomentum, self.v)

    @_lazy_property
    def periapsis(self):
        """Periapsis coordinate of orbit in meters.
        """
        import numbers
        pdist = self.p / (1 + self.e)
        if isinstance(self.a, numbers.Real):
            return normed(self.LRL) * pdist
        else:
            return normed(self.LRL) * pdist[:, None]

    @_lazy_property
    def apoapsis(self):
        """Apoapsis coordinate of orbit in meters.
        """
        import numbers
        if isinstance(self.a, numbers.Real):
            if self.a < 0:
                return np.array([np.inf, np.inf, np.inf])
            adist = self.p / (1 - self.e)
            return -normed(self.LRL) * adist
        else:
            wBound = self.a > 0
            apoapsis = np.empty((len(self.a), 3))
            if np.any(wBound):
                adist = self.p[wBound] / (1 - self.e[wBound])
                apoapsis[wBound] = -normed(self.LRL)[wBound] * adist[wBound, None]
            if np.any(~wBound):
                apoapsis[~wBound] = np.array([np.inf, np.inf, np.inf])
            return apoapsis

    @_lazy_property
    def tle(self):
        """Two line element for this Orbit.
        """
        from .io import make_tle
        return make_tle(*self.kozaiMeanKeplerianElements, self.t)


class EarthObserver:
    """ An earth-bound observer.

    Parameters
    ----------
    lon : float
        Geodetic longitude in degrees (increasing to the East).
    lat : float
        Geodetic latitude in degrees.
    elevation : float, optional
        Elevation in meters.
    fast : bool, optional
        Use fast lookup tables for Earth Orientation parameters.  ~ meter and
        ~10 micron/sec accurate for dates between approximately 1973 and the
        present.  Less accurate outside this range.

    Attributes
    ----------
    lon
    lat
    elevation
    itrs : array_like (3,)
        ITRS coordinates in meters.

    Methods
    -------
    getRV(time)
        Get position and velocity at specified time(s) in the GCRF frame.
    """
    def __init__(self, lon, lat, elevation=0, fast=False):
        from astropy.coordinates import EarthLocation
        import astropy.units as u

        self.lon = lon
        self.lat = lat
        self.elevation = elevation
        self._location = EarthLocation(
            self.lon * u.deg, self.lat * u.deg, self.elevation * u.m
        )
        self.itrs = self._location.itrs.cartesian.xyz.to(u.m).value
        self.fast = fast

    def __repr__(self):
        out = "EarthObserver(lon={!r}, lat={!r}".format(self.lon, self.lat)
        if self.elevation != 0:
            out += ", elevation={!r}".format(self.elevation)
        out += ")"
        return out

    _dGHAdt = (
        np.array([[0, 1, 0], [-1, 0, 0], [0, 0, 0]], dtype=float) * 1.002737909350795 * 2 * np.pi / 86400
    )

    def getRV(self, time):
        """ Get position and velocity at specified time(s).

        Parameters
        ----------
        time : array_like or astropy.time.Time (n,)
            If float (array), then should correspond to GPS seconds;
            i.e., seconds since 1980-01-06 00:00:00 UTC

        Returns
        -------
        r : array_like (n, 3)
            Position in meters.
        v : array_like (n, 3)
            Velocity in meters per second.

        Notes
        -----
        The returned position and velocity are in the GCRF frame.
        """
        import astropy.units as u

        if self.fast:
            import numbers
            if isinstance(time, astropy.time.Time):
                time = time.gps
            if isinstance(time, numbers.Real):
                scalar = True
                time = np.array([time])
            else:
                scalar = False
            mjd_tt = _gpsToTT(time)
            d_ut1_tt_mjd, pmx, pmy = iers_interp(time)
            pn = erfa.pnm80(2400000.5, mjd_tt)
            gst = erfa.gst94(2400000.5, mjd_tt + d_ut1_tt_mjd)
            cg, sg = np.cos(gst), np.sin(gst)
            gstMat = np.zeros([len(time), 3, 3], dtype=float)
            gstMat[:, 0, 0] = cg
            gstMat[:, 0, 1] = sg
            gstMat[:, 1, 0] = -sg
            gstMat[:, 1, 1] = cg
            gstMat[:, 2, 2] = 1.0

            polar = np.zeros([len(time), 3, 3], dtype=float)
            polar[:] = np.eye(3)

            polar[:, 0, 2] = pmx
            polar[:, 1, 2] = -pmy
            polar[:, 2, 0] = -pmx
            polar[:, 2, 1] = pmy

            U = gstMat @ pn
            outR = np.transpose(polar @ U, (0, 2, 1)) @ self.itrs
            outV = np.transpose(polar @ self._dGHAdt @ U, (0, 2, 1)) @ self.itrs

            if scalar:
                return outR[0], outV[0]
            return outR, outV
        else:
            if not isinstance(time, astropy.time.Time):
                time = astropy.time.Time(time, format='gps')
            r, v = self._location.get_gcrs_posvel(time)
            return r.xyz.to(u.m).T.value, v.xyz.to(u.m / u.s).T.value

    def sunAlt(self, time):
        """Get Sun altitude for observer at `time`.

        Parameters
        ----------
        time : array_like or astropy.time.Time (n,)
            If float (array), then should correspond to GPS seconds;
            i.e., seconds since 1980-01-06 00:00:00 UTC

        Returns
        -------
        alt : array_like(n,)
            Altitude of sun in radians.
        """
        if isinstance(time, astropy.time.Time):
            time = time.gps
        ro, _ = self.getRV(time)
        r_sun = sunPos(time, fast=False)
        dr = r_sun - ro
        return np.pi / 2 - unitAngle3(normed(ro), normed(dr))


class OrbitalObserver:
    """ An observer in orbit.

    Parameters
    ----------
    orbit : Orbit
        The orbit of the observer.
    propagator : Propagator, optional
        The propagator instance to use.

    Attributes
    ----------
    orbit
    propagator

    Methods
    -------
    getRV(time)
        Get position and velocity at specified time(s).
    """
    def __init__(self, orbit, propagator=_KeplerianPropagator()):
        self.orbit = orbit
        self.propagator = propagator

    def __repr__(self):
        out = "OrbitalObserver({!r}".format(self.orbit)
        if self.propagator != _KeplerianPropagator():
            out += ", propagator={!r}".format(self.propagator)
        out += ")"
        return out

    def getRV(self, time):
        """ Get position and velocity at specified time(s).

        Parameters
        ----------
        time : array_like or astropy.time.Time (n,)
            If float (array), then should correspond to GPS seconds;
            i.e., seconds since 1980-01-06 00:00:00 UTC

        Returns
        -------
        r : array_like (n, 3)
            Position in meters.
        v : array_like (n, 3)
            Velocity in meters per second.
        """
        from .compute import rv
        return rv(self.orbit, time, propagator=self.propagator)



import pytest
import numpy as np

from astropy.time import Time
import astropy.units as u
from astropy.coordinates import Longitude, Latitude
from astropy.table import QTable

import ssapy
import ssapy.utils as utils
from .ssapy_test_helpers import timer
from ssapy.constants import RGEO, VGEO


@pytest.mark.slow
@timer
def test_particles():
    np.random.seed(562)

    time0 = Time(2458316., format='jd')

    # Setup observer location
    lon = 100.0
    lat = 33.0
    elevation = 1300.0
    earthObserver = ssapy.EarthObserver(lon, lat, elevation)

    #
    # Create orbit for the target
    #
    # a = RGEO * 0.98
    # e = 0.01
    # i = 0.001
    # pa = 0.0
    # raan = 1.2
    # trueAnomaly = 2.03
    # orbit = Orbit.fromKeplerianElements(a, e, i, raan, pa, trueAnomaly, time0)

    r = np.random.uniform(RGEO*0.9, RGEO*1.1)
    theta = np.random.uniform(0, 2*np.pi)
    phi = np.random.uniform(0, np.pi)
    x = r * np.cos(theta) * np.sin(phi)
    y = r * np.sin(theta) * np.sin(phi)
    z = r * np.cos(phi)
    r = np.array([x, y, z])
    # Pick a velocity near VGEO
    v = np.random.uniform(VGEO*0.9, VGEO*1.1)
    # Pick a random direction
    theta = np.random.uniform(0, 2*np.pi)
    phi = np.random.uniform(0, np.pi)
    vx = v * np.cos(theta) * np.sin(phi)
    vy = v * np.sin(theta) * np.sin(phi)
    vz = v * np.cos(phi)
    v = np.array([vx, vy, vz])
    orbit = ssapy.Orbit(r, v, time0)

    num_epochs = 2
    particles = []
    epoch_times = []
    lnprobs = []
    ch_rmean = []
    ch_vmean = []
    time = time0
    for _ in range(num_epochs):
        epoch_times.append(time)

        # Build observation times
        earthTimes = time + np.linspace(0, 10, 5)*u.h
        earthRaDec = ssapy.radec(orbit, earthTimes, observer=earthObserver)
        earthRStation, earthVStation = earthObserver.getRV(earthTimes)

        arc = QTable()
        arc['ra'] = Longitude(earthRaDec[0]*u.rad)
        arc['dec'] = Latitude(earthRaDec[1]*u.rad)
        arc['rStation_GCRF'] = earthRStation*u.m
        arc['vStation_GCRF'] = earthVStation*u.m/u.s
        arc['time'] = Time(earthTimes)
        arc['sigma'] = np.ones(5, dtype=float)*u.arcsec

        # initialize somewhere near the truth, but perturbed a bit and with large uncertainties
        initializer = ssapy.GaussianRVInitializer(r, v, rSigma=0.1*RGEO, vSigma=0.1*VGEO)
        # Use R and A priors
        priors = [ssapy.rvsampler.RPrior(RGEO, RGEO*0.2), ssapy.rvsampler.APrior(RGEO, RGEO*0.2)]

        # Run MCMC sampler to generate particles for the observed epoch
        rvprob = ssapy.RVProbability(arc, time, priors=priors)
        sampler = ssapy.EmceeSampler(rvprob, initializer, nWalker=50)
        chain, lnprob, lnprior = sampler.sample(nBurn=500, nStep=100)
        chain, lnprob, lnprior = ssapy.utils.cluster_emcee_walkers(chain, lnprob, lnprior, thresh_multiplier=4)

        ch_rmean.append(np.mean(chain[...,0:3], axis=(0,1)))
        ch_vmean.append(np.mean(chain[...,3:6], axis=(0,1)))

        # Create Particles object from sampler output
        particles.append(ssapy.Particles(chain, rvprob, lnpriors=lnprior))
        lnprobs.append(lnprob.ravel())

        time += 12.0 * u.h
    epoch_times = Time(epoch_times)
    # Check that epochs in particles class match the input epoch times
    for i in range(num_epochs):
        np.testing.assert_almost_equal(particles[i].epoch, epoch_times[i].gps)

    # Check that particles likelihood matches that stored from Sampler.
    # There is some offset here because of the prior in Sampler that is not used in the
    # Particles method.
    for i in range(num_epochs):
        lnP = particles[i].lnlike(particles[i].orbits) + particles[i].lnpriors
        np.testing.assert_allclose(lnP, lnprobs[i])

    # Check that mean particle values in each epoch match reference values
    rref, vref = ssapy.rv(orbit, epoch_times)

    for i in range(num_epochs):
        pmean = particles[i].mean()
        rmean = pmean[0:3]
        vmean = pmean[3:6]

        print("--------------------")
        print("r difference wrt GEO")
        # print((ch_rmean[i]-rref[i])/RGEO)
        print((rmean - rref[i]) / RGEO)
        print("v difference wrt GEO")
        # print((ch_vmean[i]-vref[i])/VGEO)
        print((vmean - vref[i]) / VGEO)
        np.testing.assert_allclose(rmean, rref[i], rtol=0, atol=RGEO*0.001)
        np.testing.assert_allclose(vmean, vref[i], rtol=0, atol=VGEO*0.001)

    # Check that Particles.fuse method gives correct mean of particle values
    particles[0].fuse(particles[1])
    pmean = particles[0].mean()
    rmean = pmean[0:3]
    vmean = pmean[3:6]
    print("--------------------")
    print("   AFTER PARTICLE FUSION   ")
    print("r difference wrt GEO")
    print((rmean - rref[i]) / RGEO)
    print("v difference wrt GEO")
    print((vmean - vref[i]) / VGEO)
    np.testing.assert_allclose(rmean, rref[0], rtol=0, atol=RGEO*0.001)
    np.testing.assert_allclose(vmean, vref[0], rtol=0, atol=VGEO*0.001)


if __name__ == '__main__':
    test_particles()



from setuptools import setup, Extension, find_packages
from setuptools.command.build_ext import build_ext
from setuptools.command.install import install
import subprocess
import sys
import os

class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=''):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)

class CMakeBuild(build_ext):
    def run(self):
        try:
            _ = subprocess.check_output(['cmake', '--version'])
        except OSError:
            raise RuntimeError("CMake must be installed to build the following extensions: " + ", ".join(e.name for e in self.extensions))

        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext):
        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
        cmake_args = ['-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=' + extdir,
                      '-DPYTHON_EXECUTABLE=' + sys.executable]

        cfg = 'Debug' if self.debug else 'Release'
        build_args = ['--config', cfg]

        cmake_args += ['-DCMAKE_BUILD_TYPE=' + cfg]
        build_args += ['--', '-j4']

        env = os.environ.copy()
        env['CXXFLAGS'] = '{} -DVERSION_INFO=\\"{}\\"'.format(env.get('CXXFLAGS', ''),
                                                              self.distribution.get_version())
        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)
        if 'CMAKE_VERBOSE_MAKEFILE' in env:
            cmake_args += ['-DCMAKE_VERBOSE_MAKEFILE=1']
        subprocess.check_call(['cmake', ext.sourcedir] + cmake_args, cwd=self.build_temp, env=env)
        subprocess.check_call(['cmake', '--build', '.'] + build_args, cwd=self.build_temp)


setup(
    name='ssapy',
    version='1.0.0',
    ext_modules=[CMakeExtension("ssapy._ssapy")],
    cmdclass={"build_ext": CMakeBuild},
    packages=find_packages(),
    package_data={'ssapy': ['_ssapy*.so']},  # Adjust as necessary
    license='MIT',
    zip_safe=False,
    include_package_data=True,
)




# flake8: noqa: E501

"""
Functions to simplify certain aspects of SSAPy. 
For when you want a quick answer and do not need want to use high fidelity defaults.
"""

from .utils import get_times, points_on_circle
from . import Orbit, rv
from .accel import AccelKepler, AccelSolRad, AccelEarthRad, AccelDrag
from .body import get_body
from .gravity import AccelHarmonic, AccelThirdBody
from .propagator import RK78Propagator

from astropy.time import Time
import numpy as np
from typing import Union, List, Tuple


def keplerian_prop(integration_timestep: float = 10) -> RK78Propagator:
    """
    Create and return an RK78Propagator for a Keplerian orbit.

    This propagator uses only the Keplerian acceleration for a two-body problem.

    Parameters:
    ----------
    integration_timestep : float, optional
        The time step for the RK78Propagator integration (default is 10 seconds).

    Returns:
    -------
    RK78Propagator
        An instance of RK78Propagator configured with Keplerian acceleration.
    """
    return RK78Propagator(AccelKepler(), h=integration_timestep)


accel_3_cache = None
def threebody_prop(integration_timestep: float = 10) -> RK78Propagator:
    """
    Create and return an RK78Propagator with a set of accelerations for a three-body problem.

    The three bodies considered are Earth (Keplerian effect), Moon, and the Earth itself.

    Parameters:
    ----------
    integration_timestep : float, optional
        The time step for the RK78Propagator integration (default is 10 seconds).

    Returns:
    -------
    RK78Propagator
        An instance of RK78Propagator configured with the three-body accelerations.
    """
    global accel_3_cache
    if accel_3_cache is None:
        accel_3_cache = AccelKepler() + AccelThirdBody(get_body("moon"))
    return RK78Propagator(accel_3_cache, h=integration_timestep)


accel_4_cache = None
def fourbody_prop(integration_timestep: float = 10) -> RK78Propagator:
    """
    Create and return an RK78Propagator with a set of accelerations for a four-body problem.

    The four bodies considered are Earth (Keplerian effect), Moon, Sun, and the Earth itself.

    Parameters:
    ----------
    integration_timestep : float, optional
        The time step for the RK78Propagator integration (default is 10 seconds).

    Returns:
    -------
    RK78Propagator
        An instance of RK78Propagator configured with the four-body accelerations.
    """
    global accel_4_cache
    if accel_4_cache is None:
        accel_4_cache = AccelKepler() + AccelThirdBody(get_body("moon")) + AccelThirdBody(get_body("Sun"))
    return RK78Propagator(accel_4_cache, h=integration_timestep)


accel_best_cache = None
def best_prop(integration_timestep=10, kwargs=dict(mass=250, area=.022, CD=2.3, CR=1.3)):
    """
    Create and return an RK78Propagator with a comprehensive set of accelerations.

    Parameters:
    ----------
    integration_timestep : float, optional
        The time step for the RK78Propagator integration (default is 10 seconds).
    kwargs : dict, optional
        Dictionary of parameters for non-conservative forces (mass, area, CD, CR).
        If not provided, defaults are used.

    Returns:
    -------
    RK78Propagator
        An instance of RK78Propagator configured with cached accelerations.
    """
    global accel_best_cache
    if accel_best_cache is None:
        aEarth = AccelKepler() + AccelHarmonic(get_body("Earth", model="EGM2008"), 140, 140)
        aMoon = AccelThirdBody(get_body("moon")) + AccelHarmonic(get_body("moon"), 20, 20)
        aSun = AccelThirdBody(get_body("Sun"))
        aMercury = AccelThirdBody(get_body("Mercury"))
        aVenus = AccelThirdBody(get_body("Venus"))
        aMars = AccelThirdBody(get_body("Mars"))
        aJupiter = AccelThirdBody(get_body("Jupiter"))
        aSaturn = AccelThirdBody(get_body("Saturn"))
        aUranus = AccelThirdBody(get_body("Uranus"))
        aNeptune = AccelThirdBody(get_body("Neptune"))
        nonConservative = AccelSolRad(**kwargs) + AccelEarthRad(**kwargs) + AccelDrag(**kwargs)
        planets = aMercury + aVenus + aMars + aJupiter + aSaturn + aUranus + aNeptune
        accel_best_cache = aEarth + aMoon + aSun + planets + nonConservative
    return RK78Propagator(accel_best_cache, h=integration_timestep)


def ssapy_kwargs(mass=250, area=0.022, CD=2.3, CR=1.3):
    """
    Generate a dictionary of default parameters for a space object used in simulations.

    Parameters:
    ----------
    mass : float, optional
        Mass of the object in kilograms (default is 250 kg).
    area : float, optional
        Cross-sectional area of the object in square meters (default is 0.022 m^2).
    CD : float, optional
        Drag coefficient of the object (default is 2.3).
    CR : float, optional
        Radiation pressure coefficient of the object (default is 1.3).

    Returns:
    -------
    dict
        A dictionary containing the parameters for the space object.
    """
    # Asteroid parameters
    kwargs = dict(
        mass=mass,  # [kg]
        area=area,  # [m^2]
        CD=CD,  # Drag coefficient
        CR=CR,  # Radiation pressure coefficient
    )
    return kwargs


def ssapy_prop(integration_timestep=60, propkw=ssapy_kwargs()):
    """
    Setup and return an RK78 propagator with specified accelerations and radiation pressure effects.

    Parameters:
    ----------
    integration_timestep : int
        Time step for the numerical integration (in seconds).
    propkw : dict, optional
        Keyword arguments for radiation pressure accelerations. If None, default arguments are used.

    Returns:
    -------
    RK78Propagator
        An RK78 propagator configured with the specified accelerations and time step.
    """
    # Accelerations - pass a body object or string of body name.
    moon = get_body("moon")
    sun = get_body("Sun")
    Mercury = get_body("Mercury")
    Venus = get_body("Venus")
    Earth = get_body("Earth", model="EGM2008")
    Mars = get_body("Mars")
    Jupiter = get_body("Jupiter")
    Saturn = get_body("Saturn")
    Uranus = get_body("Uranus")
    Neptune = get_body("Neptune")
    aEarth = AccelKepler() + AccelHarmonic(Earth, 140, 140)
    aSun = AccelThirdBody(sun)
    aMoon = AccelThirdBody(moon) + AccelHarmonic(moon, 20, 20)
    aSolRad = AccelSolRad(**propkw)
    aEarthRad = AccelEarthRad(**propkw)
    accel = aEarth + aMoon + aSun + aSolRad + aEarthRad
    # Build propagator
    prop = RK78Propagator(accel, h=integration_timestep)
    return prop


# Uses the current best propagator and acceleration models in ssapy
def ssapy_orbit(orbit=None, a=None, e=0, i=0, pa=0, raan=0, ta=0, r=None, v=None, duration=(30, 'day'), freq=(1, 'hr'), t0=Time("2025-01-01", scale='utc'), t=None, prop=ssapy_prop()):
    """
    Compute the orbit of a spacecraft given either Keplerian elements or position and velocity vectors.

    Parameters:
    - orbit (Orbit, optional): An Orbit object if you already have an orbit defined.
    - a (float, optional): Semi-major axis of the orbit in meters.
    - e (float, optional): Eccentricity of the orbit (default is 0, i.e., circular orbit).
    - i (float, optional): Inclination of the orbit in degrees.
    - pa (float, optional): Argument of perigee in degrees.
    - raan (float, optional): Right ascension of the ascending node in degrees.
    - ta (float, optional): True anomaly in degrees.
    - r (array-like, optional): Position vector in meters.
    - v (array-like, optional): Velocity vector in meters per second.
    - duration (tuple, optional): Duration of the simulation as a tuple (value, unit), where unit is 'day', 'hour', etc. Default is 30 days.
    - freq (tuple, optional): Frequency of the output as a tuple (value, unit), where unit is 'day', 'hour', etc. Default is 1 hour.
    - t0 (str, optional): Start date of the simulation in 'YYYY-MM-DD' format. Default is "2025-01-01".
    - t (array-like, optional): Specific times at which to compute the orbit. If None, times will be generated based on duration and frequency.
    - prop (function, optional): A function to compute the perturbation effects. Default is `ssapy_prop()`.

    Returns:
    - r (array-like): Position vectors of the spacecraft at the specified times.
    - v (array-like): Velocity vectors of the spacecraft at the specified times.
    - t (array-like): Times at which the orbit was computed. Returned only if `t` was None.

    Raises:
    - ValueError: If neither Keplerian elements nor position and velocity vectors are provided.
    - RuntimeError or ValueError: If an error occurs during computation.
    """
    t0 = Time(t0, scale='utc')
    if t is None:
        time_is_None = True
        t = get_times(duration=duration, freq=freq, t0=t0)
    else:
        t0 = t[0]
        time_is_None = False

    if orbit is not None:
        pass
    elif a is not None:
        kElements = [a, e, i, pa, raan, ta]
        orbit = Orbit.fromKeplerianElements(*kElements, t0)
    elif r is not None and v is not None:
        orbit = Orbit(r, v, t0)
    else:
        raise ValueError("Either Keplerian elements (a, e, i, pa, raan, ta) or position and velocity vectors (r, v) must be provided.")

    try:
        r, v = rv(orbit, t, prop)
        if time_is_None:
            return r, v, t
        else:
            return r, v
    except (RuntimeError, ValueError) as err:
        print(err)
        return np.nan, np.nan, np.nan


def get_similar_orbits(r0, v0, rad=1e5, num_orbits=4, duration=(90, 'days'), freq=(1, 'hour'), start_date="2025-1-1", mass=250):
    """
    Generate similar orbits by varying the initial position.

    Parameters:
    ----------
    r0 : array_like
        Initial position vector of shape (3,).
    v0 : array_like
        Initial velocity vector of shape (3,).
    rad : float
        Radius of the circle around the initial position to generate similar orbits.
    num_orbits : int
        Number of similar orbits to generate.
    duration : tuple
        Duration of the orbit simulation.
    freq : tuple
        Frequency of output data.
    start_date : str
        Start date for the simulation.
    mass : float
        Mass of the satellite.

    Returns:
    -------
    trajectories : ndarray
        Stacked array of shape (3, n_times, num_orbits) containing the trajectories.
    """
    r0 = np.reshape(r0, (1, 3))
    v0 = np.reshape(v0, (1, 3))
    print(r0, v0)
    for idx, point in enumerate(points_on_circle(r0, v0, rad=rad, num_points=num_orbits)):
        # Calculate entire satellite trajectory
        r, v = ssapy_orbit(r=point, v=v0, duration=duration, freq=freq, start_date=start_date, integration_timestep=10, mass=mass, area=mass / 19000 + 0.01, CD=2.3, CR=1.3)
        if idx == 0:
            trajectories = np.concatenate((r0, v0), axis=1)[:len(r)]
        rv = np.concatenate((r, v), axis=1)
        trajectories = np.dstack((trajectories, rv))
    return trajectories
